#include "declarations.h"
#include <unistd.h>
#include <errno.h>

void compile_callback(GdsFile *current)
{
   gchar *path;
   gchar output[384];
   gchar *line[1];

   if(feof(current->compile_file))
   {
      pclose(current->compile_file);
      gdk_input_remove(current->compile_id);
      current->compile_id = 0;
      current->compile_file = NULL;
      g_print("Input callback removed.\n");
      return;
   }

   while(fgets(output, sizeof(output), current->compile_file))
   {
      if(path = strrchr(output, '\n')) *path = '\0';
      line[0] = output;
      gtk_clist_append(GTK_CLIST(current->compile_list), line);
   }
   return;
}

void compile_file(GtkWidget *widget, gpointer data)
{
   gchar *string = NULL;
   gchar output[384];
   gchar *path = NULL;
   GdsFile *current = NULL;

   if(cur_file->compile_id != 0) return;
   path = get_path_from_filename(cur_file->filename);
   if(cur_file->props.dir) chdir(cur_file->props.dir);
   else chdir(path);
   g_free(path);
   current = cur_file;
   if(cur_file->props.compiler)
   {
      if(gds_editor_changed(GDS_EDITOR(cur_file->text)))
      {
         file_not_saved(cur_file->filename, cur_file, GTK_SIGNAL_FUNC(compile_file), GTK_SIGNAL_FUNC(compile_file));
         return;
      }
      string = replace_percent_tokens(cur_file->props.compiler);
      g_print("%s\n", string);
      gtk_clist_clear(GTK_CLIST(current->compile_list));
      if(!(current->compile_file = popen(string, "r" )))
      {
         perror("Error opening execute process!");
         return;	
      }
      current->compile_id = gdk_input_add(fileno(current->compile_file), GDK_INPUT_READ, (GdkInputFunction) compile_callback, (gpointer) current);
      g_free(string);
   }
}

void debug_file(GtkWidget *widget, gpointer data)
{
   gint pid;
   FILE *process = NULL;
   gchar *string = NULL;
   gchar output[384];
   gchar *path = NULL;
   GdsFile *current = NULL;
   gchar *line[1];
   current = cur_file;

   pid = fork();
   if(pid > 0)
   {
      current->debug_pid = pid;
      return;
   }
   /* we are in a new process here */
   path = get_path_from_filename(cur_file->filename);
   if(cur_file->props.dir) chdir(cur_file->props.dir);
   else chdir(path);
   g_free(path);

   if(cur_file->props.debugger)
   {
      string = replace_percent_tokens(cur_file->props.debugger);
      g_print("%s\n", string);
      if(!(process = popen(string, "r" )))
      {
         perror("Error opening debug process!");
         return;	
      }
      while(fgets(output, sizeof(output), process))
      {
         g_print("%s", output);
      }
      pclose(process);
      g_free(string);
   }
   _exit(9);
   current->debug_pid = 0;
}


void execute_file(GtkWidget *widget, gpointer data)
{
   gint pid;
   FILE *process = NULL;
   gchar *string = NULL;
   gchar output[384];
   gchar *path = NULL;
   GdsFile *current = NULL;
   gchar *line[1];
   current = cur_file;

   pid = fork();
   if(pid > 0)
   {
      current->exec_pid = pid;
      return;
   }
   /* we are in a new process here */
   path = get_path_from_filename(cur_file->filename);
   if(cur_file->props.dir) chdir(cur_file->props.dir);
   else chdir(path);
   g_free(path);

   if(cur_file->props.execution)
   {
      string = replace_percent_tokens(cur_file->props.execution);
      g_print("%s\n", string);

      if(!(process = popen(string, "r" )))
      {
         perror("Error opening execute process!");
         return;	
      }
      pclose(process);
      g_free(string);
   }
   _exit(9);
   current->exec_pid = 0;
}

void clear_list(GtkWidget *widget, gpointer data)
{
   gtk_clist_clear(GTK_CLIST(cur_file->compile_list));
}

gchar *replace_percent_tokens(const gchar *string)
{
   gchar *path;
   gchar *filename;
   gchar *file;
   gchar *new_string = NULL;
   gchar *ptr;
   path = get_path_from_filename(cur_file->filename);
   filename = get_file_from_filename(cur_file->filename);
   file = get_file_wo_extension(cur_file->filename);
   new_string = g_strdup(string);
   while(1)
   {
      ptr = g_str_replace(new_string, "%1", path);
      if(ptr) { g_free(new_string); new_string = ptr; }
      else { break; }
   }
   while(1)
   {
      ptr = g_str_replace(new_string, "%2", filename);
      if(ptr) { g_free(new_string); new_string = ptr; }
      else { break; }
   }
   while(1)
   {
      ptr = g_str_replace(new_string, "%3", file);
      if(ptr) { g_free(new_string); new_string = ptr; }
      else { break; }
   }
   return(new_string);
}

gchar *get_path_from_filename(gchar *filename)
{
   gchar *file = NULL;
   gchar *text = NULL;
   gchar *value = NULL;
   file = g_strdup(filename);
   text =  strrchr(file, '/');
   if(text)
   {
      *text = '\0';
      text = file;
      value = g_strdup(text);
   }
   g_free(file);
   return(value);
}

gchar *get_file_from_filename(gchar *filename)
{
   gchar *text = NULL;
   text =  strrchr(filename, '/');
   if(text) { text++; } else { text = filename; }
   return(g_strdup(text));
}

gchar *get_file_wo_extension(gchar *filename)
{
   gchar *file = NULL;
   gchar *text = NULL;
   gchar *extension = NULL;
   gchar *value = NULL;
   file = g_strdup(filename);
   text =  strrchr(file, '/');
   if(text) { text++; } else { text = file; }
   extension = strrchr(text, '.');
   if(extension) { *extension = '\0'; }
   value =  g_strdup(text);
   g_free(file);
   return(value);
}

gchar *g_str_replace(const gchar *input, const gchar *replace, const gchar *with)
{
   gchar *text = NULL;
   gchar *start = NULL;
   gchar *end = NULL;
   gchar *value = NULL;
   text = g_strdup(input);
   start = strstr(text, replace);
   if(!start) return(NULL);
   end = start + strlen(replace);
   *start = '\0';
   value = g_strconcat(text, with, end, NULL);
   g_free(text);
   return(value);
}
