#include "declarations.h"

void undo_last(GtkWidget *widget, gpointer data)
{
   if(!gtk_extext_undo_is_empty(GTK_EXTEXT(cur_file->text)))
   {
      gtk_extext_undo(GTK_EXTEXT(cur_file->text));
      GDS_EDITOR(cur_file->text)->changed -= 2;
      gtk_signal_emit_by_name(GTK_OBJECT(cur_file->text), "move_to_column", 0, NULL);
   }
}

void redo_last(GtkWidget *widget, gpointer data)
{
   gtk_extext_redo(GTK_EXTEXT(cur_file->text));
}

void cut_text(GtkWidget *widget, gpointer data)
{
   GtkEditable *editable;
   editable = (GtkEditable *)cur_file->text;
   gtk_editable_cut_clipboard(editable);
   editable->has_selection = FALSE;
   editable->selection_start_pos = (gint)-1;
   editable->selection_end_pos = (gint)-1;
}

void copy_text(GtkWidget *widget, gpointer data)
{
   gtk_editable_copy_clipboard(GTK_EDITABLE(cur_file->text));
}

void paste_text(GtkWidget *widget, gpointer data)
{
   gtk_editable_paste_clipboard(GTK_EDITABLE(cur_file->text));
}

void clear_text(GtkWidget *widget, gpointer data)
{
   GtkEditable *editable;
   editable = GTK_EDITABLE(cur_file->text);
   if(editable->has_selection)
      gtk_editable_delete_selection(editable);
   else
   {
      select_all_text(widget, data);
      gtk_editable_delete_selection(editable);
   }
}

void clear_entry(GtkWidget *widget, gpointer data)
{
   GtkEntry *entry;
   entry = GTK_ENTRY(filename_entry);
   gtk_entry_set_text(entry, "");
}

void comment_selection(GtkWidget *widget, gpointer data)
{
   gint start, end;
   GtkEditable *editable;
   editable = GTK_EDITABLE(cur_file->text);
   if(editable->has_selection)
   {
      start = editable->selection_start_pos;
      end = editable->selection_end_pos + 2;
      editable->has_selection = FALSE;
      gtk_extext_set_position(editable, start);
      gtk_editable_insert_text(editable, "/*", 2, (gint *)&editable->current_pos);
      gtk_extext_set_position(editable, end);
      gtk_editable_insert_text(editable, "*/", 2, (gint *)&editable->current_pos);
      gtk_editable_select_region(editable, start, end + 2);
   }
}

void select_current(GtkWidget *widget, gpointer data)
{
   gint start, end;
   GtkEditable *editable;
   editable = GTK_EDITABLE(cur_file->text);
   start = gtk_editable_get_position(editable);
   if(gtk_extext_get_current_word(GTK_EXTEXT(editable), &start, &end))
      gtk_editable_select_region(editable, start, end);
}

void select_all_text(GtkWidget *widget, gpointer data)
{
   gtk_editable_select_region(GTK_EDITABLE(cur_file->text), 0, gtk_extext_get_length(GTK_EXTEXT(cur_file->text)));
}

void copy_selection_to_command_entry(GtkWidget *widget, gpointer data)
{
   GtkEditable *editable;
   gchar *text;
   gint start, end;

   editable = GTK_EDITABLE(cur_file->text);
   
   if(!editable->has_selection)
   {
      start = gtk_editable_get_position(editable);
      if(gtk_extext_get_current_word(GTK_EXTEXT(editable), &start, &end))
         gtk_editable_select_region(editable, start, end);
   }
   if(editable->has_selection)
   {
      text = gtk_editable_get_chars(editable, editable->selection_start_pos, editable->selection_end_pos);
      gtk_entry_set_text(GTK_ENTRY(filename_entry), text);
      g_free(text);
   }
   gtk_widget_grab_focus(filename_entry);
}

void highlight_syntax(GtkWidget *widget, gpointer data)
{
   GtkExText *extext;
   extext = (GtkExText *)cur_file->text;
   gds_editor_highlight(GDS_EDITOR(extext));
}
