#include "declarations.h"
#include <unistd.h>

gchar *last_dir = NULL;
GList *recent_files = NULL;

void new_window(GtkWidget *widget, gchar *filename)
{
   gint pid;
   gchar *command = NULL;
   pid = fork();
   if(pid > 0) return;
   command = g_strconcat(PACKAGE, filename, NULL);
   system(command);
   g_free(command);
   _exit(9);
}

void new_file(GtkWidget *widget, gpointer data)
{
   gchar *string;
   gchar *entry_text;
   entry_text = gtk_entry_get_text(GTK_ENTRY(filename_entry));
   if(strlen(entry_text) > 0)
   {
      gtk_entry_set_text(GTK_ENTRY(filename_entry), "");
      string =  g_strdup(entry_text);
      strcpy(string, entry_text);
   }
   else   
   {
      string = g_new(char, 35);
      g_snprintf(string, 35, _("Untitled%d"), new_num);
      new_num++;
   }
   cur_file = create_file(string);
   g_free(string);
   attach_highlight_tables(cur_file);
   display_message("New file created.", FLASH);
   adjust_sensitivity();
}

void open_file(GtkWidget *widget, gpointer data)
{
   file_selector = last_dir ? gnome_filelist_new_with_path(last_dir) : gnome_filelist_new_with_path(general_preferences.default_dir);
   gnome_filelist_set_title(GNOME_FILELIST(file_selector), _("Open File..."));
   gtk_signal_connect(GTK_OBJECT(GNOME_FILELIST(file_selector)->ok_button), "clicked", GTK_SIGNAL_FUNC(file_ok_sel), file_selector);
   gtk_signal_connect_object(GTK_OBJECT(GNOME_FILELIST(file_selector)->ok_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(file_selector));
   gtk_signal_connect_object(GTK_OBJECT(GNOME_FILELIST(file_selector)->cancel_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(file_selector));
   gtk_widget_show(file_selector);
}

void file_ok_sel(GtkWidget *widget, GnomeFileList *file_list)
{
   gchar *string;   
   string = gnome_filelist_get_filename(GNOME_FILELIST(file_list));
   if(check_if_file_exists(string))
   {
      finish_file_ok_sel(widget, string);
   }
   else
   {
      if(check_if_file_can_exist(string))
      {
         remove(string);
         file_dne_error(string, GTK_SIGNAL_FUNC(finish_file_ok_sel));
      }
      else
      {
         read_error(string);
      }
   }
   g_free(string);
}

void finish_file_ok_sel(GtkWidget *widget, gchar *filename)
{
   gchar *temp = NULL;
   temp = get_path_from_filename(filename);
   if(temp && strcmp(filename, temp) != 0)
   {
      g_free(last_dir);
      last_dir = g_new(char, strlen(temp)+1);
      strcpy(last_dir, temp);
   }
   g_free(temp);
   cur_file = create_file(filename);
   check_if_read_only(filename, cur_file, TRUE, TRUE);
   open_file_add_text(cur_file->text, filename);
   add_file_to_recent_list(filename);
   attach_highlight_tables(cur_file);
   if(cur_file->tables) copy_properties(&cur_file->props, &cur_file->tables->props);
   cur_file->props.dir = get_path_from_filename(cur_file->filename);
   gtk_extext_undo_clear_all(GTK_EXTEXT(cur_file->text));
   GDS_EDITOR(cur_file->text)->changed = FALSE;
   adjust_sensitivity();
   display_message("File opened.", FLASH);
}

void revert_current_file(GtkWidget *widget, gpointer data)
{
   if(!GDS_EDITOR(cur_file->text)->changed) return;
   if(check_if_file_exists(cur_file->filename))
   {
      gtk_editable_delete_text(GTK_EDITABLE(cur_file->text), 0, gtk_extext_get_length(GTK_EXTEXT(cur_file->text)));
      open_file_add_text(cur_file->text, cur_file->filename);
      GDS_EDITOR(cur_file->text)->changed = FALSE;
      display_message("Reverted.", FLASH);
   }
}

void save_file(GtkWidget *widget, gpointer data)
{
   FILE *file;
   gchar *all_text;   
   file = fopen(cur_file->filename, "w");
   if(file == NULL)
   {
      save_error(cur_file->filename);
      return;   
   }
   all_text = gtk_editable_get_chars(GTK_EDITABLE(cur_file->text), 0, gtk_extext_get_length(GTK_EXTEXT(cur_file->text)));
   fprintf(file, "%s", all_text);
   fclose(file);
   GDS_EDITOR(cur_file->text)->changed = FALSE;
   gtk_signal_emit_by_name(GTK_OBJECT(cur_file->text), "move_to_column", 0, NULL);
   display_message("File Saved.", FLASH);
}

void save_file_as(GtkWidget *widget, gpointer data)
{
   gchar *str = NULL;
   str = get_path_from_filename(cur_file->filename);
   file_selector = gnome_filelist_new_with_path(str);
   gnome_filelist_set_title(GNOME_FILELIST(file_selector), _("Save file with new name..."));
   gtk_signal_connect(GTK_OBJECT(GNOME_FILELIST(file_selector)->ok_button), "clicked", GTK_SIGNAL_FUNC(save_as_file), file_selector);
   gtk_signal_connect_object(GTK_OBJECT(GNOME_FILELIST(file_selector)->ok_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(file_selector));
   gtk_signal_connect_object(GTK_OBJECT(GNOME_FILELIST(file_selector)->cancel_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(file_selector));
   gtk_widget_show(file_selector);
   g_free(str);
}

void save_as_file(GtkWidget *widget, GnomeFileList *file_list)
{
   gchar *string;
   string = gnome_filelist_get_filename(GNOME_FILELIST(file_list));
   if(check_if_file_exists(string))
      file_exists_error(string, GTK_SIGNAL_FUNC(finish_save_as_file));
   else
      finish_save_as_file(widget, string);
   g_free(string);
}

void finish_save_as_file(GtkWidget *widget, gchar *filename)
{
   FILE *file;
   gchar *all_text;   

   if(!(file = fopen(filename, "w")))
   {
      save_error(filename);
      return;   
   }
   all_text = gtk_editable_get_chars(GTK_EDITABLE(cur_file->text), 0, gtk_extext_get_length(GTK_EXTEXT(cur_file->text)));
   fprintf(file, "%s", all_text);
   fclose(file);

   gds_file_set_filename(cur_file, filename);
   gds_file_set_title(cur_file, filename);

   gtk_editable_set_editable(GTK_EDITABLE(cur_file->text), TRUE);
   add_file_to_recent_list(filename);
   attach_highlight_tables(cur_file);
   if(cur_file->tables && !cur_file->props.over_ride) copy_properties(&cur_file->props, &cur_file->tables->props);
   cur_file->read_only = FALSE;
   GDS_EDITOR(cur_file->text)->changed = FALSE;
   gtk_signal_emit_by_name(GTK_OBJECT(cur_file->text), "move_to_column", 0, NULL);
   set_window_title();
   adjust_sensitivity();
   display_message("File saved with new name.", FLASH);
}

void save_all_files(GtkWidget *widget, gpointer data)
{
   GdsFile *file;
   GList *temp;
   gint backup;
   gint i = 0;
   backup = cur_page;
   display_message("Saving all open files...", FLASH);   
   for(i = total_files-1; i >= 0; i--)
   {
      file = (GdsFile *)g_list_nth_data(files_list, i);
      file_change(widget, file);
      save_file(NULL, NULL);
   }
   gtk_notebook_set_page(GTK_NOTEBOOK(files_book), backup);
   display_message("All files saved.", FLASH);   
}

void close_file(GtkWidget *widget, gpointer data)
{
   if(gds_editor_changed(GDS_EDITOR(cur_file->text)))
   {
      file_not_saved(cur_file->filename, cur_file, NULL, NULL);
   }
   else
   {
      close_file_ok(widget, data);
   }
}

void close_file_ok(GtkWidget *widget, gpointer data)
{
   close_file_full(widget, cur_file);
}

void close_file_full(GtkWidget *widget, GdsFile *file)
{
   gint old_page;
   if(GDS_TEAR_AWAY(cur_file->tear_away)->torn)
      gds_tear_away_bust_a_switch(GDS_TEAR_AWAY(cur_file->tear_away));
   detach_highlight_tables(file);
   files_list = g_list_remove(files_list, file);
   total_files--;
   if(!total_files) { files_list = NULL; }
   while(gtk_events_pending()) gtk_main_iteration();
   old_page = cur_page;
   gtk_widget_destroy(GTK_WIDGET(file));
   if(old_page == 0) { cur_page = old_page; }
   gtk_notebook_set_page(GTK_NOTEBOOK(files_book), cur_page);
   set_notebook_page(GTK_NOTEBOOK(files_book), NULL, cur_page, NULL);
   adjust_sensitivity();
   display_message("File closed.", FLASH);   
}

void close_all_files(GtkWidget *widget, gpointer data)
{
   GdsFile *file;
   GList *temp;
   gint i = 0;
   
   for(i = total_files-1; i >= 0; i--)
   {
      file = (GdsFile *)g_list_nth_data(files_list, i);
      file_change(widget, file);
      if(gds_editor_changed(GDS_EDITOR(file->text)))
      {
         file_not_saved(file->filename, file, GTK_SIGNAL_FUNC(try_exit), GTK_SIGNAL_FUNC(close_all_files));
         return;
      }
      else
      {
         close_file_ok(widget, data);
      }
   }
}

gboolean open_file_add_text(GtkWidget *text, gchar *filename)
{
#define BUF_SIZE 32768
   FILE *file;
   gpointer buffer;
   gint bytes;
   gint pos;
   pos = gtk_extext_get_position(GTK_EDITABLE(text));
   
   if(!(file = fopen(filename, "rb")))
   {
      return FALSE;
   }
   gtk_extext_freeze(GTK_EXTEXT(text));
   buffer = g_new(char, BUF_SIZE);
   while(bytes = fread(buffer, sizeof(char), BUF_SIZE, file))
   {
      gtk_extext_insert(GTK_EXTEXT(text), (gchar *)buffer, bytes);
      if(bytes < BUF_SIZE)
         break;
   }
   g_free(buffer);   
   fclose(file);
   gtk_extext_set_position(GTK_EDITABLE(text), pos);
   gtk_extext_thaw(GTK_EXTEXT(text));
   gtk_widget_grab_focus(text);
   return TRUE;
}

gboolean check_if_file_exists(gchar *filename)
{
   gboolean success;
   FILE *file;
   if(!(file = fopen(filename, "r")))
   {
      success = FALSE;
   }
   else
   {
      success = TRUE;
      fclose(file);
   }
   return(success);
}

gboolean check_if_file_can_exist(gchar *filename)
{
   gboolean success;
   FILE *file;
   if(!(file = fopen(filename, "a")))
   {
      success = FALSE;
   }
   else
   {
      success = TRUE;
      fclose(file);
   }
   return(success);
}

gboolean check_if_read_only(gchar *filename, GdsFile *temp, gboolean set_ro, gboolean do_error)
{
   FILE *file;
   if(!(file = fopen(filename, "a"))) //we arent going to edit it, just goint to check if we could...
   {
      if(set_ro)
      {
         gtk_editable_set_editable(GTK_EDITABLE(temp->text), FALSE);
         temp->read_only = TRUE;
      }
      if(do_error)
         read_only(filename); // just let em know that they cant edit it :o)
   }
   else
   {
      if(set_ro)
      {
         gtk_editable_set_editable(GTK_EDITABLE(temp->text), TRUE);
         temp->read_only = FALSE;
      }
      fclose(file);
   }
}

void file_create(GtkWidget *widget, gchar *filename)
{
   FILE *file;
   if(!(file = fopen(filename, "w"))) //create it if it doesnt exist
      fclose(file);
}

GList *build_file_list()
{
   GList *temp;
   temp = g_list_copy(files_list);
   return(temp);
}

void setup_recent_files_list(gint max_files)
{
   GList *temp = NULL;
   GList *current = NULL;
   temp = build_glist_from_file("files.hist", max_files);
   build_recent_files_menu(temp);
   
   current = g_list_first(temp);
   while(current)
   {
      g_free(current->data);
      current = current->next;
   }
   g_list_free(temp);
}

void build_recent_files_menu(GList *file_strings)
{
   GList *current = NULL;
   gchar *temp = NULL;
   GtkWidget *menuitem;
   
   kill_old_recent_menu();
   
   current = g_list_first(file_strings);
   while(current)
   {
      temp = (gchar *)current->data;
      menuitem = gtk_menu_item_new_with_label(temp);
      recent_files = g_list_append(recent_files, (gpointer)menuitem);
      gtk_signal_connect(GTK_OBJECT(menuitem), "activate", GTK_SIGNAL_FUNC(recent_files_callback), 0);  
      gtk_menu_append(GTK_MENU(recentmenu), menuitem);
      gtk_widget_show(menuitem);
      current = current->next;
   }
}

void recent_files_callback(GtkWidget *widget, gpointer data)
{
   GtkWidget *accel;
   gchar *string;
   g_return_if_fail(GTK_IS_BIN(widget));
   
   accel = GTK_BIN(widget)->child;
   gtk_label_get(GTK_LABEL(accel), &string);

   if(check_if_file_exists(string))
   {
      finish_file_ok_sel(widget, string);
   }
   else
   {
      if(check_if_file_can_exist(string))
      {
         remove(string);
         file_dne_error(string, GTK_SIGNAL_FUNC(finish_file_ok_sel));
      }
      else
      {
         read_error(string);
      }
   }
}

void add_file_to_recent_list(gchar *filename)
{
   add_string_to_file("files.hist", filename);
   setup_recent_files_list(general_preferences.history);
}

void kill_old_recent_menu()
{
   GList *temp = NULL;
   GtkWidget *current;
   
   temp = g_list_first(recent_files);
   while(temp)
   {
      current = (GtkWidget *)temp->data;
      if(GTK_IS_WIDGET(current))
         gtk_widget_destroy(current);
      temp = temp->next;
   }
   g_list_free(recent_files);
   recent_files = NULL;
}

