#include "declarations.h"
void main_window_focused(GtkWidget *widget, gpointer data);
void session_die(GnomeClient *client, gpointer data);
static gint session_manage_save(GnomeClient *client, gint phase, GnomeSaveStyle save_style, gint is_shutdown, GnomeInteractStyle interact_style, gint is_fast, gpointer client_data);

GtkWidget *app;
GtkWidget *window;
GtkWidget *appbar;
GtkWidget *whole_window_box;
GtkWidget *main_window_box;
GtkWidget *box;
GtkWidget *switch_label;
GtkWidget *paned;
GtkWidget *edit_box;
GtkWidget *files_book;
GtkWidget *file_selector;
GtkStyle *extext_style = NULL;
GtkAccelGroup *accel_group;
GList *files_list = NULL;
GdsFile *cur_file = NULL;
gint new_num = 0;
gint cur_page = -1;
gint total_files = 0;
static gchar *input_file = NULL, *input_session = NULL;
gint *line = NULL;
gint files_opened = 0;
GnomeClient *client;


struct poptOption options[] =
{
   { 
      "session", 's', POPT_ARG_STRING, &input_session,
       0, N_("Specify the session file to use."), NULL
   },
   {
      "file", 'f', POPT_ARG_STRING, &input_file,
      0, N_("Open a specific file."), NULL
   },
   {
      "line", 'l', POPT_ARG_INT, &line,
      0, N_("Jump to a line in the file specified by -f."),
      NULL
   },
   {
      NULL, '\0', 0, NULL, 0, NULL, NULL
   }
};

int main(int argc, char *argv[])
{
   poptContext pctx;
   gchar** args;
   init_colors();
   directory_check();
   
   /* do some gettext stuff */
   bindtextdomain(PACKAGE, GNOMELOCALEDIR);
   textdomain(PACKAGE);

   /* initialize the window and give it a name and a size */

   gnome_init_with_popt_table(PACKAGE, PACKAGE " v" VERSION, argc, argv, options, 0, &pctx);
   args = (gchar **)poptGetArgs(pctx);
   app = gnome_app_new(PACKAGE, VERSION);
   gtk_container_set_border_width(GTK_CONTAINER(app), 5);
   gtk_window_set_policy(GTK_WINDOW(app), FALSE, TRUE, FALSE);
   gtk_signal_connect(GTK_OBJECT(app), "focus_in_event", GTK_SIGNAL_FUNC(main_window_focused), NULL);
   gtk_signal_connect(GTK_OBJECT(app), "delete_event", GTK_SIGNAL_FUNC(pre_exit), NULL);
   window = app;
   
   appbar = gnome_appbar_new(FALSE, TRUE, GNOME_PREFERENCES_ALWAYS);
   gnome_appbar_set_default(GNOME_APPBAR(appbar), "Ready.");
   gnome_app_set_statusbar(GNOME_APP(app), appbar);
   gtk_widget_show(appbar);

   make_menus();
   make_toolbar();

/* What to show when showing nothing else; defaults to nothing */
   whole_window_box = gtk_vbox_new(FALSE, 0);
   gtk_container_set_border_width(GTK_CONTAINER(whole_window_box), 0);
   gnome_app_set_contents(GNOME_APP(window), whole_window_box);
   gtk_widget_show(whole_window_box);

   main_window_box = gtk_hbox_new(FALSE, 0);
   gtk_container_set_border_width(GTK_CONTAINER(main_window_box), 0);
   gtk_box_pack_start(GTK_BOX(whole_window_box), main_window_box, TRUE, TRUE, 0);
   gtk_widget_show(main_window_box);

   edit_box = gtk_hbox_new(FALSE, 0);
   gtk_container_set_border_width(GTK_CONTAINER(edit_box), 2);
   gtk_container_add(GTK_CONTAINER(main_window_box), edit_box);
   gtk_widget_show(edit_box);

   files_book = gtk_notebook_new();
   gtk_notebook_set_tab_pos(GTK_NOTEBOOK(files_book), GTK_POS_TOP);
   gtk_box_pack_end(GTK_BOX(edit_box), files_book, TRUE, TRUE, 0);
   gtk_notebook_set_scrollable(GTK_NOTEBOOK(files_book), TRUE);
   gtk_notebook_popup_enable(GTK_NOTEBOOK(files_book));
   gtk_signal_connect(GTK_OBJECT(files_book), "switch_page", GTK_SIGNAL_FUNC(set_notebook_page), NULL);  
   gtk_widget_show(files_book);

   get_prefs();

   if(general_preferences.notebook_tab_position >= 4)
   {
      gtk_notebook_set_show_tabs(GTK_NOTEBOOK(files_book), FALSE);
   }
   else
   {
      gtk_notebook_set_show_tabs(GTK_NOTEBOOK(files_book), TRUE);
      gtk_notebook_set_tab_pos(GTK_NOTEBOOK(files_book), (GtkPositionType) general_preferences.notebook_tab_position);
   }

   extext_style = gtk_style_copy(gtk_widget_get_style(window));
   gtk_style_ref(extext_style);
   extext_style->text[GTK_STATE_NORMAL]= general_preferences.fg;
   extext_style->bg[GTK_STATE_NORMAL]= general_preferences.bg;
   extext_style->text[GTK_STATE_SELECTED]= general_preferences.s_fg;
   extext_style->bg[GTK_STATE_SELECTED]= general_preferences.s_bg;
   extext_style->base[GTK_STATE_NORMAL]= general_preferences.bg;
   extext_style->bg_pixmap[0] = NULL;
   extext_style->bg_pixmap[1] = NULL;
   extext_style->bg_pixmap[2] = NULL;
   extext_style->bg_pixmap[3] = NULL;
   extext_style->bg_pixmap[4] = NULL;
   general_preferences.font = gdk_font_load(general_preferences.font_name);
   if(general_preferences.font)
   {
      extext_style->font = general_preferences.font;
      gdk_font_ref(general_preferences.font);
   }

   if(input_session)
   {
      if(check_if_file_exists(input_session))
      {
         files_opened = open_session_from_file(input_session);
         if(files_opened)
         {
            gchar *path, *file, *full, cwd[256];
            g_free(session_file);
            getcwd(cwd, sizeof(cwd));
            path = get_path_from_filename(input_session);
            file = get_file_from_filename(input_session);
            if(!path) full = g_strconcat(cwd, "/", file, NULL);
            else full = g_strconcat(path, "/", file, NULL);
            g_free(path);
            g_free(file);
            session_file = full;
         }
         else
         {
            g_error("Session file: %s could not be opened, or contained no session data.\n", input_session);
         }
      }
   }
   if(input_file)
   {
      if(check_if_file_exists(input_file))
      {
         finish_file_ok_sel(file_menu_open, input_file);
         if(line)
         {
   	    if(*line > 0) *line--;
            if(*line <= GTK_EXTEXT(cur_file->text)->line_count && *line > 0)
            {
               gtk_extext_set_line(GTK_EXTEXT(cur_file->text), *line);
            }
         }
         files_opened++;
      }
      else
      {
         g_error("File: %s could not be opened, or does not exist.\n", input_file);
      }
   }
   if(!files_opened && args)
   {
      int i = 0;
      while(args[i])
      {
         if(check_if_file_exists((gchar *)args[i]))
         {
            finish_file_ok_sel(file_menu_open, (gchar *)args[i]);
            files_opened++;
         }
         i++;
      }
   }
   if(!files_opened)
   {
      if(!get_last_session())
         new_file(file_menu_new, NULL);
   }
   poptFreeContext(pctx);

   client = gnome_master_client();
   gtk_signal_connect(GTK_OBJECT(client), "save_yourself", GTK_SIGNAL_FUNC(session_manage_save), argv[0]);
   gtk_signal_connect(GTK_OBJECT(client), "die", GTK_SIGNAL_FUNC(session_die), argv[0]);
   setup_recent_files_list(general_preferences.history);
   gtk_widget_show(window);
   set_window_title();
   adjust_sensitivity();
   gtk_main();
   return 0;
}


void file_change(GtkWidget *widget, GdsFile *new_file)
{
   cur_file = new_file;
   if(cur_file)
   {
      if(GDS_TEAR_AWAY(cur_file->tear_away)->torn)
      {
         gdk_window_raise(GDS_TEAR_AWAY(cur_file->tear_away)->new_parent->window);
      }
      else
      {
         gtk_notebook_set_page(GTK_NOTEBOOK(files_book), get_nth_file_number(cur_file));
         gdk_window_raise(window->window);
      }
   }
   set_window_title();
   adjust_sensitivity();
}

void set_notebook_page(GtkNotebook *notebook, GtkNotebookPage *page, guint page_num, gpointer data)
{
   cur_page = page_num;
   cur_file = get_nth_file_proper(page_num);
   if(cur_file) { gtk_widget_grab_focus(cur_file->text); }
   file_pos_changed(NULL, 0, NULL);
   set_window_title();
   adjust_sensitivity();   
}

void set_window_title()
{
   gchar *string;
   char buffer[378] = PACKAGE " v" VERSION;

   if(cur_file)
   {
      strcat(buffer, " Session: ");
      string =  get_file_from_filename(session_file);
      strcat(buffer, string);
      g_free(string);
      strcat(buffer, " File: ");
      string =  get_file_from_filename(cur_file->filename);
      strcat(buffer, string);
      g_free(string);
   }
   gtk_window_set_title(&GNOME_APP(window)->parent_object, buffer);
}


void file_pos_changed(GtkWidget *widget, gint pos, gpointer data)
{
   gchar line_string[128];
   gchar changed[56];
   gint length;
   gint line;
   gint lines;
   gint col;
   guchar cur_char;
   if(cur_file != NULL)
   {
      col = gtk_extext_get_column(GTK_EXTEXT(cur_file->text));   
      pos = gtk_extext_get_position(GTK_EDITABLE(cur_file->text));
      length = gtk_extext_get_length(GTK_EXTEXT(cur_file->text));
      line = gtk_extext_get_line(GTK_EXTEXT(cur_file->text)) + 1;
      lines = GTK_EXTEXT(cur_file->text)->line_count + 1;
      cur_char = gtk_extext_get_char_at_pos(GTK_EXTEXT(cur_file->text), pos);
      g_snprintf(line_string, sizeof(line_string), _("Line: %d / %d Col: %d Byte: %d / %d Value: %d"), line, lines, col, pos, length, cur_char);
      if(gds_editor_changed(GDS_EDITOR(cur_file->text))) strcat(line_string, _("  (changed)"));
      gtk_entry_set_text(GTK_ENTRY(cur_file->file_data), line_string);
   }
}

void adjust_sensitivity()
{
   if(cur_file != NULL)
   {
      gtk_widget_set_sensitive(file_menu_revert, TRUE);
      gtk_widget_set_sensitive(file_menu_close, TRUE);
      gtk_widget_set_sensitive(main_toolbar_close, TRUE);   
      gtk_widget_set_sensitive(file_menu_close_all, TRUE);
      gtk_widget_set_sensitive(file_menu_save, TRUE);
      gtk_widget_set_sensitive(main_toolbar_save, TRUE);   
      gtk_widget_set_sensitive(file_menu_saveas, TRUE);
      gtk_widget_set_sensitive(main_toolbar_saveas, TRUE);   
      gtk_widget_set_sensitive(file_menu_save_all, TRUE);

      gtk_widget_set_sensitive(editmenu, TRUE);
      gtk_widget_set_sensitive(searchmenu, TRUE);
      gtk_widget_set_sensitive(buildmenu, TRUE);
      gtk_widget_set_sensitive(insertmenu, TRUE);

      gtk_widget_set_sensitive(main_toolbar_undo, TRUE);
      gtk_widget_set_sensitive(main_toolbar_redo, TRUE);
      gtk_widget_set_sensitive(main_toolbar_cut, TRUE);
      gtk_widget_set_sensitive(main_toolbar_copy, TRUE);
      gtk_widget_set_sensitive(main_toolbar_paste, TRUE);
   }
   else
   {
      gtk_widget_set_sensitive(file_menu_revert, FALSE);
      gtk_widget_set_sensitive(file_menu_close, FALSE);
      gtk_widget_set_sensitive(file_menu_close_all, FALSE);
      gtk_widget_set_sensitive(main_toolbar_close, FALSE);   
      gtk_widget_set_sensitive(file_menu_save, FALSE);
      gtk_widget_set_sensitive(main_toolbar_save, FALSE);   
      gtk_widget_set_sensitive(file_menu_saveas, FALSE);
      gtk_widget_set_sensitive(main_toolbar_saveas, FALSE);   
      gtk_widget_set_sensitive(file_menu_save_all, FALSE);
      
      gtk_widget_set_sensitive(editmenu, FALSE);
      gtk_widget_set_sensitive(searchmenu, FALSE);
      gtk_widget_set_sensitive(buildmenu, FALSE);
      gtk_widget_set_sensitive(insertmenu, FALSE);
      
      gtk_widget_set_sensitive(main_toolbar_undo, FALSE);
      gtk_widget_set_sensitive(main_toolbar_redo, FALSE);
      gtk_widget_set_sensitive(main_toolbar_cut, FALSE);
      gtk_widget_set_sensitive(main_toolbar_copy, FALSE);
      gtk_widget_set_sensitive(main_toolbar_paste, FALSE);
   }
}


void directory_check()
{
   gchar buffer[384];
   g_snprintf(buffer, sizeof(buffer), "%s/." PACKAGE "/session.last", getenv("HOME"));
   session_file = g_strdup(buffer);
   gboolean exists = check_if_file_exists(buffer);
   if(exists) return;
   g_snprintf(buffer, sizeof(buffer), "mkdir %s/." PACKAGE, getenv("HOME"));
   g_print(_("Creating ~/" PACKAGE " directory.\n"));
   system(buffer);

   g_snprintf(buffer, sizeof(buffer), "%s/." PACKAGE "/session.last", getenv("HOME"));
   g_print(_("Creating ~/." PACKAGE "/session.last file.\n"));
   file_create(NULL, buffer);

   g_snprintf(buffer, sizeof(buffer), "mkdir %s/." PACKAGE "/languages", getenv("HOME"));
   g_print(_("Creating ~/." PACKAGE "/languages directory.\n"));
   system(buffer);
   return;
}

void kill_widget(GtkWidget *widget, GtkWidget *target)
{
   gtk_widget_destroy(GTK_WIDGET(target));
}

gint delete_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
   return(TRUE);
}

void destroy(GtkWidget *widget, gpointer data)
{
   gtk_main_quit();
}

void pre_exit(GtkWidget *widget, gpointer data)
{
   if(file_selector) { gtk_widget_destroy(file_selector); }
   if(general_preferences.save_session) { save_session_cb(NULL, NULL); }
   try_exit(NULL, NULL);
}

void try_exit(GtkWidget *widget, gpointer data)
{
   GdsFile *file;
   GList *temp;
   gint i = 0;
   
   for(i = total_files-1; i >= 0; i--)
   {
      file = (GdsFile *)g_list_nth_data(files_list, i);
      file_change(widget, file);
      if(gds_editor_changed(GDS_EDITOR(file->text)))
      {
         file_not_saved(file->filename, file, GTK_SIGNAL_FUNC(try_exit), GTK_SIGNAL_FUNC(try_exit));
         return;
      }
      else
      {
         close_file_ok(widget, data);
      }
   }
   /* basic cleanup...probably not necessary, but im anal like that */
   gtk_style_unref(extext_style);
   if(general_preferences.font)
      gdk_font_unref(general_preferences.font);
   g_free(last_dir);
   g_free(session_file);
   kill_old_recent_menu();
   /* end anal-ness */   
   gtk_main_quit();
}


GdsFile *get_nth_file_proper(gint number)
{
   GdsFile *set = NULL;
   gint iterations = 0;
   gint current = -1;
   while(iterations < total_files)
   {
      set = (GdsFile *)g_list_nth_data(files_list, iterations);
      if(!set) break;
      if(!GDS_TEAR_AWAY(set->tear_away)->torn) current++;
      if(current == number) break;
      iterations++;
   }
   return(set);
}

gint get_nth_file_number(GdsFile *gds_file) // zero is bad, so we go from 1 to the end instead
{
   GdsFile *check = NULL;
   gint iterations = 0;
   gint current = 0;
   while(iterations < total_files)
   {
      check = (GdsFile *)g_list_nth_data(files_list, iterations);
      if(check == gds_file) { current = iterations; break; }
      if(!check) break;
      if(!GDS_TEAR_AWAY(check->tear_away)->torn) current++;
      iterations++;
   }
   return(current);
}

void main_window_focused(GtkWidget *widget, gpointer data)
{
   gint iterations = 0;
   guint total_pages;
   GdsFile *new_file = NULL;
   total_pages = g_list_length(GTK_NOTEBOOK(files_book)->children); // max page number
   if(cur_page >= 0 && cur_page < total_pages-1)
      new_file = get_nth_file_proper(cur_page);
   else
      new_file = get_nth_file_proper(total_pages-1);
   file_change(widget, new_file);
}

void display_message(const gchar *text, gint type)
{
   if(type == FLASH)
      gnome_app_flash(GNOME_APP(window), text);
   if(type == ERROR)
      gnome_app_error(GNOME_APP(window), text);
   if(type == WARNING)
      gnome_app_warning(GNOME_APP(window), text);
}

void session_die(GnomeClient *client, gpointer data)
{
   gtk_main_quit();
}

static gint session_manage_save(GnomeClient *client, gint phase, GnomeSaveStyle save_style, gint is_shutdown, GnomeInteractStyle interact_style, gint is_fast, gpointer client_data)
{
   gchar **argv;
   guint argc;
   gchar rm[] = "rm";
   gchar *options[] = { "-s", "-f" };
   gchar buffer[384];

   g_snprintf(buffer, sizeof(buffer), "%s/." PACKAGE, getenv("HOME"));
   gnome_client_set_current_directory(client, buffer);
   strcat(buffer, "/");
   strcat(buffer, client->client_id);
   save_session_to_file(buffer);

   argv = (gchar **)g_malloc0(sizeof(gchar *) * 4);
   argc = 3;
   argv[0] = (gchar *)client_data;
   argv[1] = options[0];
   argv[2] = client->client_id;
   gnome_client_set_clone_command(client, argc, argv);
   gnome_client_set_restart_command(client, argc, argv);
   argv[0] = rm;
   argv[1] = options[1];
   argv[2] = client->client_id;
   gnome_client_set_discard_command(client, argc, argv);
   g_free(argv);
   return TRUE;
}
