#include "declarations.h"

gint check_list_for_string(GList *list, gchar *string);

/* 
 * Find (F6)
 * If the file has text in the filename_entry, look for it.
 * Elsewise we need to popup the dialog.
 */ 
void do_find(GtkWidget *widget, gpointer data)
{
   gchar *text;
   gboolean found;
   GtkExTextMatch match;
   GtkEditable *editable;
   g_return_if_fail(cur_file != NULL);
  
   if(get_entry_length(GTK_ENTRY(filename_entry)) > 0)
   {
      text = get_entry_text(GTK_ENTRY(filename_entry));
      editable = GTK_EDITABLE(cur_file->text);
      if(editable->has_selection && get_selection_length(editable) > 48)
         found = find_text_in_file(GTK_EXTEXT(cur_file->text), editable->selection_start_pos, editable->selection_end_pos, text, TRUE, &match);
      else
         found = find_text_in_file(GTK_EXTEXT(cur_file->text), gtk_extext_get_position(GTK_EDITABLE(cur_file->text)), gtk_extext_get_length(GTK_EXTEXT(cur_file->text)), text, TRUE, &match);
      if(!found)
         search_not_found(text);
      g_free(text);
   }
   else
   {
      widget = gnome_search_dialog_new(GNOME_SEARCH_FIND);
      gtk_signal_connect(GTK_OBJECT(GNOME_SEARCH_DIALOG(widget)->find_button), "clicked", GTK_SIGNAL_FUNC(find_text), widget);
      gtk_signal_connect_object(GTK_OBJECT(GNOME_SEARCH_DIALOG(widget)->cancel_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(widget));
      gnome_search_dialog_set_find_history(GNOME_SEARCH_DIALOG(widget), build_glist_from_file("find.hist", general_preferences.history));
      set_find_text_from_selection(GNOME_SEARCH_DIALOG(widget));
      gtk_widget_show(widget);
   }
}

/*
 * Find callback
 */
void find_text(GtkWidget *widget, GnomeSearchDialog *search_dialog)
{
   GtkExTextMatch match;
   gchar *search_string;
   gboolean found;
   gboolean iscase;
   gint pos;
   gint old_pos;

   g_return_if_fail(cur_file != NULL);

   old_pos = gtk_extext_get_position(GTK_EDITABLE(cur_file->text));
   search_string = gnome_search_dialog_get_find_text(GNOME_SEARCH_DIALOG(search_dialog));
   if(gnome_search_dialog_get_search_style(search_dialog) == GNOME_SEARCH_STYLE_NORM)
   {
      if(gnome_search_dialog_get_search_start(search_dialog) == GNOME_SEARCH_START_TOP)
         { pos = 0; } else { pos = old_pos; }
      gnome_search_dialog_set_startpos(search_dialog, (guint)pos);      
      iscase = gnome_search_dialog_case_sensitive(search_dialog);
      if(!search_dialog->search_selection)
         gnome_search_dialog_set_endpos(search_dialog, (guint)gtk_extext_get_length(GTK_EXTEXT(cur_file->text)));
      found = find_text_in_file(GTK_EXTEXT(cur_file->text), pos, search_dialog->endpos, search_string, iscase, &match);
      if(gnome_search_dialog_add_find_string(search_dialog, gnome_search_dialog_get_find_text(search_dialog)))
         add_string_to_file("find.hist", gnome_search_dialog_get_find_text(search_dialog));
      if(found)
      {
         if(gnome_search_dialog_get_dialog_type(search_dialog) == GNOME_SEARCH_FIND && !gnome_search_dialog_keep_window_open(search_dialog))
   	 {
            gtk_widget_destroy(GTK_WIDGET(search_dialog));
   	 }
      }
      else
      {
         search_not_found(gnome_search_dialog_get_find_text(search_dialog));
      }
   }
   else
   {
      return;
   }
}

/*
 * Find some text in a GtkExText with the given settings
 */
gboolean find_text_in_file(GtkExText *extext, gint start, gint end, gchar *search_string, gboolean case_sensitive, GtkExTextMatch *match)
{
   gboolean found = FALSE;
   g_return_val_if_fail(extext != NULL, FALSE);
   found = gtk_extext_search(extext, search_string, start, case_sensitive, TRUE, match);
   if(found && match->endpos <= end)
   {
      gtk_extext_set_position(GTK_EDITABLE(extext), match->endpos);
      gtk_editable_select_region(GTK_EDITABLE(extext), match->startpos, match->endpos);
      gtk_widget_grab_focus(GTK_WIDGET(extext));
   }
   else
   {
      gtk_extext_set_position(GTK_EDITABLE(extext), start);
      found = FALSE;      
   }
   return(found);
}

/* 
 * Replace (F7)
 * If the file has a selection and text in the filename_entry, replace the text.
 * Elsewise we need to popup the dialog.
 */ 
void do_replace(GtkWidget *widget, gpointer data)
{
   gchar *text;
   g_return_if_fail(cur_file != NULL);
   if(get_entry_length(GTK_ENTRY(filename_entry)) > 0 && get_selection_length(GTK_EDITABLE(cur_file->text)) != -1)
   {
      text = get_entry_text(GTK_ENTRY(filename_entry));
      replace_text(GTK_EDITABLE(cur_file->text), text);
      g_free(text);
   }
   else
   {
      widget = gnome_search_dialog_new(GNOME_SEARCH_REPLACE);
      gtk_signal_connect(GTK_OBJECT(GNOME_SEARCH_DIALOG(widget)->find_button), "clicked", GTK_SIGNAL_FUNC(find_text), widget);
      gtk_signal_connect(GTK_OBJECT(GNOME_SEARCH_DIALOG(widget)->replace_button), "clicked", GTK_SIGNAL_FUNC(find_text_replace), widget);
      gtk_signal_connect(GTK_OBJECT(GNOME_SEARCH_DIALOG(widget)->replace_all), "clicked", GTK_SIGNAL_FUNC(replace_all_text), widget);
      gtk_signal_connect_object(GTK_OBJECT(GNOME_SEARCH_DIALOG(widget)->cancel_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(widget));
      gnome_search_dialog_set_find_history(GNOME_SEARCH_DIALOG(widget), build_glist_from_file("find.hist", general_preferences.history));
      gnome_search_dialog_set_replace_history(GNOME_SEARCH_DIALOG(widget), build_glist_from_file("replace.hist", general_preferences.history));

      set_find_text_from_selection(GNOME_SEARCH_DIALOG(widget));
      gtk_widget_show(widget);
   }
}

/*
 * Replace callback
 */
void find_text_replace(GtkWidget *widget, GnomeSearchDialog *search_dialog)
{
   gchar *find_text;
   gchar *new_text;
   gboolean found = FALSE;
   g_return_if_fail(cur_file != NULL);

   find_text = gnome_search_dialog_get_find_text(search_dialog);
   new_text = gnome_search_dialog_get_replace_text(search_dialog);

   found = replace_text(GTK_EDITABLE(cur_file->text), gnome_search_dialog_get_replace_text(search_dialog));

   search_dialog->endpos += (strlen(new_text) - strlen(find_text));

   if(gnome_search_dialog_add_replace_string(search_dialog, gnome_search_dialog_get_replace_text(search_dialog)))
      add_string_to_file("replace.hist", gnome_search_dialog_get_replace_text(search_dialog));
   if(found && !gnome_search_dialog_keep_window_open(search_dialog))
   {
      gtk_widget_destroy(GTK_WIDGET(search_dialog));
   }
}

/*
 * Replace All callback
 */
void replace_all_text(GtkWidget *widget, GnomeSearchDialog *search_dialog)
{
   guint replacements = 0;
   gint pos;
   gboolean iscase;
   gchar *find_text;
   gchar *new_text;
   GtkExTextMatch match;
   g_return_if_fail(cur_file != NULL);

   find_text = gnome_search_dialog_get_find_text(search_dialog);
   new_text = gnome_search_dialog_get_replace_text(search_dialog);
   
   if(search_dialog->search_selection)
      gtk_extext_set_position(GTK_EDITABLE(cur_file->text), search_dialog->startpos);
   else if(gnome_search_dialog_get_search_start(search_dialog) == GNOME_SEARCH_START_TOP)
      gtk_extext_set_position(GTK_EDITABLE(cur_file->text), 0);

   iscase = gnome_search_dialog_case_sensitive(search_dialog);
   while(find_text_in_file(GTK_EXTEXT(cur_file->text), gtk_extext_get_position(GTK_EDITABLE(cur_file->text)), search_dialog->endpos, find_text, iscase, &match))
   {
      replace_text(GTK_EDITABLE(cur_file->text), new_text);
      /* if we are going to get all of them, make sure to add the differences to the endpos */
      search_dialog->endpos += (strlen(new_text) - strlen(find_text));
      replacements++;
   }

   if(replacements > 0)
   {
      gchar text[128];
      if(gnome_search_dialog_add_find_string(search_dialog, gnome_search_dialog_get_find_text(search_dialog)))
         add_string_to_file("find.hist", gnome_search_dialog_get_find_text(search_dialog));
      if(gnome_search_dialog_add_replace_string(search_dialog, gnome_search_dialog_get_replace_text(search_dialog)))
         add_string_to_file("replace.hist", gnome_search_dialog_get_replace_text(search_dialog));
      if(!gnome_search_dialog_keep_window_open(search_dialog))
      {
         gtk_widget_destroy(GTK_WIDGET(search_dialog));
      }
      g_snprintf(text, sizeof(text), "%d replacements made.", replacements);
      generic_dialog_with_text(text);
   }
   else
   {
      search_not_found(gnome_search_dialog_get_find_text(search_dialog));
   }
}

/*
 * Replace the current selection in editable with replace_text
 */
gboolean replace_text(GtkEditable *editable, gchar *replace_text)
{
   gint start;
   if(editable->has_selection)
   {
      start = editable->selection_start_pos;
      gtk_editable_delete_selection(editable);
      gtk_extext_set_position(editable, start);
      gtk_extext_insert(GTK_EXTEXT(editable), replace_text, strlen(replace_text));
      return(TRUE);
   }
   else
   {
      return(FALSE);
   }
}

/*
 * Lets hop down to some line in the file.
 */
void do_gotoline(GtkWidget *widget, gpointer data)
{
   g_return_if_fail(cur_file != NULL);
   widget = gnome_goto_line_dialog_new();
   gtk_signal_connect(GTK_OBJECT(GNOME_GOTO_LINE_DIALOG(widget)->line_entry), "activate", GTK_SIGNAL_FUNC(find_line), widget);
   gtk_signal_connect(GTK_OBJECT(GNOME_GOTO_LINE_DIALOG(widget)->goto_button), "clicked", GTK_SIGNAL_FUNC(find_line), widget);
   gtk_signal_connect_object(GTK_OBJECT(GNOME_GOTO_LINE_DIALOG(widget)->cancel_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(widget));
   gtk_widget_show(widget);
   gtk_widget_grab_focus(GNOME_GOTO_LINE_DIALOG(widget)->line_entry);
}

void find_line(GtkWidget *widget, gpointer data)
{
   gint line_num;
   g_return_if_fail(cur_file != NULL);
   line_num = gnome_goto_line_dialog_get_line(GNOME_GOTO_LINE_DIALOG(data));
   if(line_num != -1 && line_num <= GTK_EXTEXT(cur_file->text)->line_count)
   {
      gtk_extext_set_line(GTK_EXTEXT(cur_file->text), line_num);
      gtk_widget_destroy(GTK_WIDGET(data));
   }
   return;
}

/* make a new process with and xterm and a manpage */
void add_manpage_string(GtkWidget *widget, gpointer data)
{
   GdsManpage *gds_manpage;
   GtkEntry *entry;
   gds_manpage = (GdsManpage *)data;
   entry = (GtkEntry *)GTK_COMBO(gds_manpage->search_combo)->entry;

   if(gds_manpage_add_search_string(gds_manpage, gtk_entry_get_text(entry)))
      add_string_to_file("man.hist", gtk_entry_get_text(entry));
}

void do_manpage(GtkWidget *widget, gpointer data)
{
   GtkEditable *editable;
   FILE *fp;
   gchar *text = NULL;
   gint start, end;
   GtkWidget *gds_manpage;

   editable = GTK_EDITABLE(cur_file->text);
   
   if(get_entry_length(GTK_ENTRY(filename_entry)) > 0)
   {
      text = gtk_entry_get_text(GTK_ENTRY(filename_entry));
   }
   else if(editable->has_selection)
   {
      text = gtk_editable_get_chars(editable, editable->selection_start_pos, editable->selection_end_pos);
   }
   else
   {
      start = gtk_editable_get_position(GTK_EDITABLE(cur_file->text));
      if(gtk_extext_get_current_word(GTK_EXTEXT(cur_file->text), &start, &end))
      {
         text = gtk_editable_get_chars(editable, start, end);
      }
   }
   gds_manpage = gds_manpage_new();
   gtk_signal_connect(GTK_OBJECT(GDS_MANPAGE(gds_manpage)->search_button), "clicked", GTK_SIGNAL_FUNC(add_manpage_string), gds_manpage);
   gds_manpage_set_history(GDS_MANPAGE(gds_manpage), build_glist_from_file("man.hist", general_preferences.history));
   gds_manpage_set_search_string(GDS_MANPAGE(gds_manpage), text);
   gtk_widget_show(gds_manpage);
   if(text)
      gtk_widget_activate(GDS_MANPAGE(gds_manpage)->search_button);
   g_free(text);
}

void set_find_text_from_selection(GnomeSearchDialog *sd)
{
   guint sel_start;
   guint sel_end;
   gchar *sel_text;

   if(GTK_EDITABLE(cur_file->text)->has_selection)
   {
      sel_start = GTK_EDITABLE(cur_file->text)->selection_start_pos;
      sel_end = GTK_EDITABLE(cur_file->text)->selection_end_pos;
      if(sel_end-sel_start <= 48)
      {
         sel_text = gtk_editable_get_chars(GTK_EDITABLE(cur_file->text), sel_start, sel_end);
         gnome_search_dialog_set_find_text(sd, sel_text);         
         gnome_search_dialog_search_selection(sd, FALSE);
         g_free(sel_text);
      }	
      else
      {
         gnome_search_dialog_set_startpos(sd, sel_start);
         gnome_search_dialog_set_endpos(sd, sel_end);
         gnome_search_dialog_search_selection(sd, TRUE);
      }
   }
   else
   {
      gnome_search_dialog_set_endpos(sd, (guint)gtk_extext_get_length(GTK_EXTEXT(cur_file->text)));
   }
}


GList *build_glist_from_file(gchar *filename, guint max)
{
   gchar full_path[256] = "";
   FILE *file;
   gchar buffer[256];
   gpointer temp;
   GList *list = NULL;
   guint iterations = 0;
   gchar *ptr = NULL;
   g_snprintf(full_path, sizeof(full_path), "%s/." PACKAGE "/%s", getenv("HOME"), filename);

   file = fopen(full_path, "r");
   if(!file) { return(list); }

   while(fgets(buffer, sizeof(buffer), file) && iterations < max)
   {
      if(!strcmp(buffer, "\n")) { continue; }
      buffer[strlen(buffer)-1] = '\0';
      if(check_list_for_string(list, buffer)) { continue; }
      temp = g_new(char, strlen(buffer)+1);
      strcpy((gchar *)temp, buffer);
      list = g_list_append(list, temp);
      iterations++;
   }
   return(list);
}


void add_string_to_file(gchar *filename, const gchar *string)
{
   gchar full_path[256] = "";
   gchar temp_file[256] = "";
   FILE *file;
   FILE *tmp_file;
   gchar buffer[256];
   
   g_snprintf(full_path, sizeof(full_path), "%s/." PACKAGE "/%s", getenv("HOME"), filename);
   g_snprintf(temp_file, sizeof(temp_file), "%s/." PACKAGE "/%s", getenv("HOME"), "yadda.tmp");

   file = fopen(full_path, "a");
   if(!file) return;
   fclose(file);

   rename(full_path, temp_file);
   
   file = fopen(full_path, "w");
   if(!file) return;

   fprintf(file, "%s\n", string);

   tmp_file = fopen(temp_file, "r");
   if(!tmp_file) return;

   while(fgets(buffer, sizeof(buffer), tmp_file) != NULL)
   {
       fprintf(file, "%s", buffer);
   }
   fclose(file);
   fclose(tmp_file);
   remove(temp_file);
}

gint get_entry_length(GtkEntry *entry)
{
   gint length;
   gchar *text;
   text = gtk_entry_get_text(entry);
   return(strlen(text));
}

/* return g_new()'d text string...must be g_free()'d later */
gchar *get_entry_text(GtkEntry *entry)
{
   gchar *text;
   gint length;
   length = get_entry_length(entry);
   text = g_new(char, length+1);
   strcpy(text, gtk_entry_get_text(entry));
   return(text);
}

gint get_selection_length(GtkEditable *editable)
{
   gint length;
   g_return_val_if_fail(editable != NULL, -1);
   if(editable->has_selection)
   {
      length = GTK_EDITABLE(cur_file->text)->selection_end_pos - GTK_EDITABLE(cur_file->text)->selection_start_pos;
      return(length);
   }
   else
   {
      return(-1);
   }
}

gchar *get_selection_text(GtkEditable *editable)
{
   gchar *text;
   gint start;
   gint end;
   
   if(editable->has_selection)
   {
      start = editable->selection_start_pos;
      end = editable->selection_end_pos;
      text = g_new(char, (end-start)+1);
      strcpy(text, gtk_editable_get_chars(editable, start, end)); // might be a mem leak here.
      return(text);
   }
   else
   {
      return(NULL);
   }
}

/* 0 for none, + for it's position in the list - use position-1 to index*/
gint check_list_for_string(GList *list, gchar *string)
{
   GList *current = NULL;
   gint return_value = 0;
   gint iterations = 0;

   current = g_list_first(list);
   while(current)
   {
      iterations++;
      if(!strcmp(string, (gchar *)current->data))
      {
         return_value = iterations;
	 break;
      }
      current = current->next;
   }
   return(return_value);
}
