#include "declarations.h"
gchar *session_file = NULL;

guint get_last_session()
{
   guint files;
   gchar buffer[256];
   g_snprintf(buffer, sizeof(buffer), "%s/." PACKAGE "/session.last", getenv("HOME"));
   if(!general_preferences.open_session) return(0);
   files = open_session_from_file(buffer);
   return(files);
}

void save_session_cb(GtkWidget *widget, gpointer data)
{
   gint success;
   gchar buffer[256];
   g_snprintf(buffer, sizeof(buffer), "%s/." PACKAGE "/session.last", getenv("HOME"));
   success = save_session_to_file(session_file ? session_file : buffer);
   if(!success) generic_dialog_with_text(_("Session file could not be written.\nMake sure that you have write access to $HOME."));
}

void open_session_cb()
{
   gchar *path;
   path =  get_path_from_filename(session_file);
   file_selector = gnome_filelist_new_with_path(path);
   g_free(path);
   gnome_filelist_set_title(GNOME_FILELIST(file_selector), _("Open Session..."));
   gtk_signal_connect(GTK_OBJECT(GNOME_FILELIST(file_selector)->ok_button), "clicked", GTK_SIGNAL_FUNC(open_session_ok), file_selector);
   gtk_signal_connect_object(GTK_OBJECT(GNOME_FILELIST(file_selector)->ok_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(file_selector));
   gtk_signal_connect_object(GTK_OBJECT(GNOME_FILELIST(file_selector)->cancel_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(file_selector));
   gtk_widget_show(file_selector);
}

void open_session_ok(GtkWidget *widget, GnomeFileList *file_list)
{
   gchar *string, *temp;   
   guint files;
   string = gnome_filelist_get_filename(GNOME_FILELIST(file_list));
   if(check_if_file_exists(string))
   {
      files = open_session_from_file(string);
      if(!files)
      {
         generic_dialog_with_text(_("File contained no session data."));
         g_free(string);
      }	
      else
      {
         g_free(session_file);
         session_file = string;
         set_window_title();
      }
   }
   else
   {
      g_free(string);
      read_error(string);
   }
}

void save_session_as(GtkWidget *widget, gpointer data)
{
   gchar buffer[384];
   g_snprintf(buffer, sizeof(buffer), "%s/." PACKAGE "/session.last", getenv("HOME"));

   file_selector = gnome_filelist_new_with_path(buffer);
   gnome_filelist_set_title(GNOME_FILELIST(file_selector), _("Save session file..."));
   gtk_signal_connect(GTK_OBJECT(GNOME_FILELIST(file_selector)->ok_button), "clicked", GTK_SIGNAL_FUNC(save_as_session), file_selector);
   gtk_signal_connect_object(GTK_OBJECT(GNOME_FILELIST(file_selector)->ok_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(file_selector));
   gtk_signal_connect_object(GTK_OBJECT(GNOME_FILELIST(file_selector)->cancel_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(file_selector));
   gtk_widget_show(file_selector);
}

void save_as_session(GtkWidget *widget, GnomeFileList *file_list)
{
   gchar *string;
   
   string = gnome_filelist_get_filename(GNOME_FILELIST(file_list));
   if(check_if_file_exists(string))
      file_exists_error(string, GTK_SIGNAL_FUNC(finish_save_as_session));
   else
      finish_save_as_session(widget, string);
   g_free(string);
}

void finish_save_as_session(GtkWidget *widget, gchar *filename)
{
   gchar *temp;
   gboolean success;   
   success = save_session_to_file(filename);
   if(!success) generic_dialog_with_text(_("Session save failed."));
   else
   {
      g_free(session_file);
      session_file = g_new(char, strlen(filename)+1);
      strcpy(session_file, filename);
      set_window_title();
   }
}

gboolean save_session_to_file(gchar *filename)
{
   FILE *file;   
   GList *temp = NULL;
   GdsFile *current;
   
   if(file = fopen(filename, "w"))
   {
      temp = g_list_first(files_list);
      while(temp)
      {
         current = (GdsFile *)temp->data;
         fprintf(file, "<file>\n");
	 fprintf(file, "filename=%s\n", current->filename);
	 fprintf(file, "over_ride=%d\n", current->props.over_ride);
	 fprintf(file, "highlight=%d\n", current->props.syntax);
	 fprintf(file, "use_spaces=%d\n", current->props.use_spaces);
	 fprintf(file, "tab_stop=%d\n", current->props.spaces);
	 fprintf(file, "line=%d\n", gtk_extext_get_line(GTK_EXTEXT(current->text)));
         if(current->props.dir)
	 fprintf(file, "directory=%s\n", current->props.dir);
         if(current->props.compiler)
	 fprintf(file, "compiler=%s\n", current->props.compiler);
         if(current->props.debugger)
	 fprintf(file, "debugger=%s\n", current->props.debugger);
         if(current->props.execution)
	 fprintf(file, "execution=%s\n", current->props.execution);	 
         fprintf(file, "</file>\n");	 
         temp = temp->next;
      }
      fclose(file);
   }
   else
   {
      return(FALSE);
   }
   return(TRUE);
}


guint open_session_from_file(gchar *filename)
{
   FILE *file;
   gint success = 0;
   gchar *text;
   gchar buffer[1024];
   gchar file_string[] = "filename=";
   gchar directory_string[] = "directory=";
   gchar compiler_string[] = "compiler=";
   gchar debugger_string[] = "debugger=";
   gchar execution_string[] = "execution=";
   gchar *newfile = NULL;
   Properties props;
   props.dir = NULL;
   props.compiler = NULL;
   props.debugger = NULL;
   props.execution = NULL;
   gint line;
   
   if(!(file = fopen(filename, "r"))) return(0);

   if(app->window != NULL)
      gdk_window_set_cursor(app->window, gdk_cursor_new(GDK_WATCH));

   while(fgets(buffer, sizeof(buffer), file))
   {
      while(gtk_events_pending()) { gtk_main_iteration(); }
      if(strstr(buffer, "<file>")) 
      {
         while(fgets(buffer, sizeof(buffer), file) != NULL)
         {
            if(strstr(buffer, "</file>")) break;
            if(text = strrchr(buffer, '\n')) *text = '\0';
            if(text = strstr(buffer, file_string))
   	    {
                  g_free(newfile);
   	       newfile = g_strdup(text + strlen(file_string));
   	       continue;
   	    }
   	    else if(sscanf(buffer, "over_ride=%d", &props.over_ride))
   	    {
   	       continue;
   	    }
   	    else if(sscanf(buffer, "highlight=%d", &props.syntax))
   	    {
   	       continue;
   	    }
   	    else if(sscanf(buffer, "auto_indent=%d", &props.auto_indent))
   	    {
   	       continue;
   	    }
   	    else if(sscanf(buffer, "use_spaces=%d", &props.use_spaces))
   	    {
   	       continue;
   	    }
   	    else if(sscanf(buffer, "tab_stop=%d", &props.spaces))
   	    {
   	       continue;
   	    }
   	    else if(sscanf(buffer, "line=%d", &line))
   	    {
   	       continue;
   	    }
            else if(text = strstr(buffer, directory_string))
   	    {
               g_free(props.dir);
               props.dir = g_strdup(text + strlen(directory_string));
               continue;
   	    }
            else if(text = strstr(buffer, compiler_string))
   	    {
               g_free(props.compiler);
   	       props.compiler = g_strdup(text + strlen(compiler_string));
               continue;
            }
            else if(text = strstr(buffer, debugger_string))
   	    {
               g_free(props.debugger);
   	       props.debugger = g_strdup(text + strlen(debugger_string));
   	       continue;
   	    }
            else if(text = strstr(buffer, execution_string))
   	    {
               g_free(props.execution);
   	       props.execution = g_strdup(text + strlen(execution_string));
   	       continue;
   	    }
         }
         if(newfile && check_if_file_exists(newfile))
         {
            finish_file_ok_sel((GtkWidget*)file_menu_open, newfile);
            success++;
            gtk_extext_set_line(GTK_EXTEXT(cur_file->text), line);
            cur_file->props.over_ride = props.over_ride;
            cur_file->props.syntax = props.syntax;
            cur_file->props.auto_indent = props.auto_indent;
            cur_file->props.use_spaces = props.use_spaces;
            cur_file->props.spaces = props.spaces;
            if(props.dir)
            {
               g_free(cur_file->props.dir);
               cur_file->props.dir = g_strdup(props.dir);
            }
            if(props.compiler)
            {
   	       g_free(cur_file->props.compiler);
   	       cur_file->props.compiler = g_strdup(props.compiler);
            }
            if(props.debugger)
            {
               g_free(cur_file->props.debugger);
               cur_file->props.debugger = g_strdup(props.debugger);
            }
            if(props.execution)
            {
               g_free(cur_file->props.execution);
               cur_file->props.execution = g_strdup(props.execution);
            }
         }
      }
   }
   fclose(file);
   g_free(newfile);
   g_free(props.dir);
   g_free(props.compiler);
   g_free(props.debugger);
   g_free(props.execution);
   if(app->window != NULL)
      gdk_window_set_cursor(app->window, gdk_cursor_new(GDK_LEFT_PTR));
   return(success);
}
