#include "declarations.h"

GdkColor colors[16];
GList *highlight_tables = NULL;

void init_colors()
{
   colors[BLACK].red = 0x0000;
   colors[BLACK].green = 0x0000;
   colors[BLACK].blue = 0x0000;

   colors[WHITE].red = 0xFFFF;
   colors[WHITE].green = 0xFFFF;
   colors[WHITE].blue = 0xFFFF;

   colors[RED].red = 0xFFFF;
   colors[RED].green = 0x0000;
   colors[RED].blue = 0x0000;

   colors[GREEN].red = 0x0000;
   colors[GREEN].green = 0xFFFF;
   colors[GREEN].blue = 0x0000;

   colors[BLUE].red = 0x0000;
   colors[BLUE].green = 0x0000;
   colors[BLUE].blue = 0xFFFF;

   colors[LT_RED].red = 0xFFFF;
   colors[LT_RED].green = 0x0000;
   colors[LT_RED].blue = 0x0000;

   colors[LT_GREEN].red = 0x0000;
   colors[LT_GREEN].green = 0xFFFF;
   colors[LT_GREEN].blue = 0x0000;

   colors[LT_BLUE].red = 0x0000;
   colors[LT_BLUE].green = 0x0000;
   colors[LT_BLUE].blue = 0xFFFF;

   colors[DK_RED].red = 0x7FFF;
   colors[DK_RED].green = 0x0000;
   colors[DK_RED].blue = 0x0000;

   colors[DK_GREEN].red = 0x0000;
   colors[DK_GREEN].green = 0x7FFF;
   colors[DK_GREEN].blue = 0x0000;

   colors[DK_BLUE].red = 0x0000;
   colors[DK_BLUE].green = 0x0000;
   colors[DK_BLUE].blue = 0x7FFF;

   colors[YELLOW].red = (255 * 255);
   colors[YELLOW].green = (255 * 255);
   colors[YELLOW].blue = 0;

   colors[ORANGE].red = (255 * 229);
   colors[ORANGE].green = (255 * 140);
   colors[ORANGE].blue = (255 * 10);
}

GdkColor *get_color(gchar *color)
{
   if(strcmp(color, "white") == 0)
      return(&colors[WHITE]);
   if(strcmp(color, "red") == 0)
      return(&colors[RED]);
   if(strcmp(color, "green") == 0)
      return(&colors[GREEN]);
   if(strcmp(color, "blue") == 0)
      return(&colors[BLUE]);
   if(strcmp(color, "light_red") == 0)
      return(&colors[LT_RED]);
   if(strcmp(color, "light_green") == 0)
      return(&colors[LT_GREEN]);
   if(strcmp(color, "light_blue") == 0)
      return(&colors[LT_BLUE]);
   if(strcmp(color, "dark_red") == 0)
      return(&colors[DK_RED]);
   if(strcmp(color, "dark_green") == 0)
      return(&colors[DK_GREEN]);
   if(strcmp(color, "dark_blue") == 0)
      return(&colors[DK_BLUE]);
   if(strcmp(color, "yellow") == 0)
      return(&colors[YELLOW]);
   if(strcmp(color, "orange") == 0)
      return(&colors[ORANGE]);

   return(&colors[BLACK]);
}


void attach_highlight_tables(GdsFile *file_ptr)
{
   gchar home_dir[384];
   struct highlight_table *table;
   gchar *extension = NULL;
   gchar *filename = NULL;   

   g_snprintf(home_dir, sizeof(home_dir), "%s/." PACKAGE "/languages", getenv("HOME"));   
   gchar install_dir[] = LANG_DIR;

   detach_highlight_tables(file_ptr);

   filename = get_file_from_filename(file_ptr->filename);
   extension = strrchr(filename, '.');
   if(extension) extension++;
   else extension = filename;
   table = find_highlight_tables(extension, home_dir);  // try the local path first
   if(!table) { table = find_highlight_tables(extension, install_dir); }	//then the install path
   if(table)
   {
      file_ptr->tables = table;
      table->refcount++;		// very important! we will delete the table if refcount is 0
      gds_editor_install_table(GDS_EDITOR(file_ptr->text), table->syntax);
      gds_editor_install_table(GDS_EDITOR(file_ptr->text), table->pattern);
      // Can we save some code in file-io if we do this here?
      gds_editor_set_highlight(GDS_EDITOR(file_ptr->text), general_preferences.syntax);
   }
   else
   {
      gds_editor_set_highlight(GDS_EDITOR(file_ptr->text), FALSE);
   }
   g_free(filename);
}


gboolean detach_highlight_tables(GdsFile *file_ptr)
{
   if(file_ptr->tables != NULL)
   {
      file_ptr->tables->refcount--;
      delete_highlight_tables(file_ptr->tables);
      file_ptr->tables = NULL;
      GDS_EDITOR(file_ptr->text)->syntax_table = NULL;
      GDS_EDITOR(file_ptr->text)->pattern_table = NULL;
      gds_editor_set_highlight(GDS_EDITOR(file_ptr->text), FALSE);      
      return(TRUE);
   }
   else
   {
      return(FALSE);
   }
}


gboolean delete_highlight_tables(struct highlight_table *table)
{
   GList *temp = NULL;
   GList *ptr;
   struct highlight_table *current;
   if(table->refcount > 0)
   {
      return(FALSE);
   }
   if(table->syntax != NULL)
      gds_editor_table_free(table->syntax);
   if(table->pattern != NULL)
      gds_editor_table_free(table->pattern);   
   temp = g_list_find(highlight_tables, table);
   if(temp)
   {
      current = (struct highlight_table *)temp->data;
      g_free(current->filename);
      g_list_delete_all(current->extensions);
      g_free(current->props.compiler);
      g_free(current->props.debugger);
      g_free(current->props.execution);
      highlight_tables = g_list_remove(highlight_tables, table); // remove the table entry
      g_free(table);	// delete the table struct
   }
   return(TRUE);
}


struct highlight_table *find_highlight_tables(gchar *extension, gchar *cur_dir)
{
   GList *temp = NULL;
   GList *ptr = NULL;
   struct highlight_table *current;
   DIR *dir_list;
   struct dirent *dir;
   FILE *lang_file;
   gchar buffer[256];
   gchar full_path[384];
   GList *exts = NULL;
   gchar *ext = NULL;
   struct highlight_table *new_table;   
   
   temp = g_list_first(highlight_tables);
   while(temp)
   {
      current = (struct highlight_table *)temp->data;
      ptr = g_list_first(current->extensions);
      while(ptr)
      {
         if(!strcmp((gchar *)ptr->data, extension))
         {
            return(current); // its already been built, just send the addy :o)
         }
         ptr = ptr->next;
      }
      temp = temp->next;
   }

   dir_list = opendir(cur_dir);
   if(dir_list != NULL)
   {
      while((dir = readdir(dir_list)) != NULL)
      {
         if(dir->d_ino > 0)
	 {
            gboolean found = FALSE;
	    temp = g_list_first(highlight_tables);
	    while(temp)
	    {
               current = (struct highlight_table *)temp->data;
	       if(!strcmp(current->filename, buffer))
               {     
                  found = TRUE;            
                  break;
	       }
	       temp = temp->next;
	    }
            if(found)	    
               continue;
	    g_snprintf(full_path, sizeof(full_path), "%s/%s", cur_dir, dir->d_name);
            exts = get_exts_list(full_path);			// get list of extensions in "full path"
            if(!exts) { continue; } 				// if list == NULL, try another file
            ext = check_exts_list(extension, exts);		// see if one of exts matches extension
            if(ext) 						// extension found!
            {
               new_table = make_highlight_tables(full_path);	// go into the file and make some tables
               if(!new_table) { return(new_table); }		// no need for empty lists thank you!
               new_table->filename = g_new(char, strlen(dir->d_name)+1);  // save the filename
	       strcpy(new_table->filename, dir->d_name);
               new_table->extensions = exts;			// copy the GList pointer
               highlight_tables = g_list_append(highlight_tables, new_table);
               gint count = get_language_settings(full_path, &new_table->props);
	       new_table->refcount = 0;
               return(new_table);
            }
            else
	    {
	       g_list_delete_all(exts);
	    }
	 }
      }
   }
   else
   {
      return(NULL);
   }
}


GList * get_exts_list(gchar *filename)
{
   FILE *file;
   gchar buffer[256];
   const char end_char = '';
   gchar name[32];
   gchar ext[24];   
   GList *temp = NULL;
   gchar *new_ext;
   gint success = 0;
   
   if(!(file = fopen(filename, "r")))
      return(NULL);

   while(fgets(buffer, sizeof(buffer), file) != NULL)
   {
      if(strstr(buffer, "<extensions>") != NULL) 
      {
         while(fgets(buffer, sizeof(buffer), file) != NULL)
         {
            if(strstr(buffer, "</extensions>")) break;
            if(success = get_str(buffer, name, "name=", end_char, sizeof(name)))            
            {
               if(success = get_str(buffer, ext, "ext=", end_char, sizeof(ext)))            
               {
                  new_ext = g_new(char, strlen(ext)+1);
                  strcpy(new_ext, ext);
                  temp = g_list_append(temp, (gpointer)new_ext);
               }
	    }
	 }
      }
   }
   fclose(file);
   return(temp);
}

gint get_language_settings(gchar *filename, Properties *props)
{
   FILE *file;
   gint success = 0;
   gchar *text;
   gchar buffer[1024];
   gchar compiler_string[] = "compiler=";
   gchar debugger_string[] = "debugger=";
   gchar execution_string[] = "execution=";
   props->over_ride = FALSE;
   props->auto_indent = 1;
   props->use_spaces = 0;
   props->spaces = 5;
   props->dir = NULL;
   props->compiler = NULL;
   props->debugger = NULL;
   props->execution = NULL;

   if(!(file = fopen(filename, "r"))) return(0);
   while(fgets(buffer, sizeof(buffer), file))
   {
      if(buffer[0]=='#') continue;
      if(strstr(buffer, "<settings>")) 
      {
         while(fgets(buffer, sizeof(buffer), file) != NULL)
         {
            if(strstr(buffer, "</settings>")) break;

            if(text = strstr(buffer, compiler_string))
	    {
	       props->compiler = g_strdup(text + strlen(compiler_string));
               success++;
	       continue;
	    }
            else if(text = strstr(buffer, debugger_string))
	    {
	       props->debugger = g_strdup(text + strlen(debugger_string));
               success++;
	       continue;
	    }
            else if(text = strstr(buffer, execution_string))
	    {
	       props->execution = g_strdup(text + strlen(execution_string));
               success++;
	       continue;
	    }
	    else if(sscanf(buffer, "auto_indent=%d", &props->auto_indent))
	    {
               success++;
	       continue;
	    }
	    else if(sscanf(buffer, "use_spaces=%d", &props->use_spaces))
	    {
               success++;
	       continue;
	    }
	    else if(sscanf(buffer, "tab_stop=%d", &props->spaces))
	    {
               success++;
	    }
         }
      }
   }
   fclose(file);
   return(success);
}

gchar *check_exts_list(gchar *ext, GList *exts)
{
   GList *temp = NULL;
   gchar *current;
   temp = g_list_first(exts);
   while(temp)
   {
      current = (gchar *)temp->data;
      if(!strcmp(ext, current))
         return(current);
      temp = temp->next;
   }
   return(NULL);
}


struct highlight_table *make_highlight_tables(gchar *filename)
{
   FILE *file;
   GList *list = NULL;
   gchar buffer[2048];
   gchar name[32];
   gchar start[64];
   gchar end[64];
   gchar words[1536];
   gchar color[32];
   gint items;
   gchar ptr1;
   gchar ptr2;
   gboolean success = FALSE;
   const char end_char = '';
   struct highlight_table *new_table;   

   if(!(file = fopen(filename, "r")))
      return(NULL);

   new_table = g_new(highlight_table, 1);
   
   while(fgets(buffer, sizeof(buffer), file) != NULL)
   {
      if(strstr(buffer, "<syntax>") != NULL)
      {
         list = NULL;
         while(fgets(buffer, sizeof(buffer), file) != NULL)
         {
            if(strstr(buffer, "</syntax>"))
	    {
	       break;
	    }
	    else
	    {
               if(success = get_str(buffer, name, "name=", end_char, sizeof(name)))
               {
                  if(success = get_str(buffer, start, "start=", end_char, sizeof(start)))
                  {
                     if(success = get_str(buffer, end, "end=", end_char, sizeof(end)))
                     {
                        if(success = get_str(buffer, color, "color=", end_char, sizeof(color)))
                        {
                           list = gds_editor_syntax_entry_new(list, name, start, end, NULL, get_color(color), NULL);	       
                        }
		     }
		  }     				   
	       }
	    }
	 }
         new_table->syntax = gds_editor_syntax_table_new(list);
      }	 

      if(strstr(buffer, "<pattern>") != NULL)
      {
         list = NULL;
         while(fgets(buffer, sizeof(buffer), file) != NULL)
         {
            if(strstr(buffer, "</pattern>"))
	    {
	       break;
	    }
	    else
	    {
               if(success = get_str(buffer, name, "name=", end_char, sizeof(name)))
               {
                  if(success = get_str(buffer, words, "words=", end_char, sizeof(words)))
                  {
                     if(success = get_str(buffer, color, "color=", end_char, sizeof(color)))
                     {
                        list = gds_editor_pattern_entry_new(list, name, words, NULL, get_color(color), NULL);
                     }
		  }
	       }	
	    }
	 }
         new_table->pattern = gds_editor_pattern_table_new(list);
      }	 
   }
   fclose(file);
   return(new_table);
}


gboolean get_str(gchar *source, gchar *dest, const gchar search[], const gchar end, gint max_chars)
{
   char cur_char;
   char last_char = '\0';
   char *ptr;

   if((ptr = strstr(source, search)) == NULL)
      return(FALSE);

   ptr += strlen(search);
   for(int iterations = 0; iterations < max_chars; iterations++)
   {
      cur_char = *ptr;
      *dest = cur_char;

      if(cur_char == end)
      {
         *dest = '\0';
         break;
      }

      if(cur_char == '\0')
      {
         return(FALSE);
      }
      
      if(last_char == '\\' && cur_char == 'n')
      {
         dest--;
	 *dest = '\n';
      }

      if(last_char == '\\' && cur_char == 't')
      {
         dest--;
	 *dest = '\t';
      }

      last_char = cur_char;
      ptr++;
      dest++;
   }
   return(TRUE);
}


void g_list_delete_all(GList *list)
{
   GList *ptr = NULL;
   ptr = g_list_first(list);
   while(ptr)
   {
      g_free(ptr->data);
      ptr = ptr->next;
   }
   g_list_free(list);
}

void highlight_tables_font_change(GdkFont *font)
{
   GList *tables = NULL;
   GList *entry = NULL;
   struct highlight_table *table;
   GdsEditorSyntaxEntry *syntax;
   GdsEditorPatternEntry *pattern;
   
   tables = g_list_first(highlight_tables);
   while(tables)
   {
      table = (struct highlight_table *)tables->data;
      entry = g_list_first(table->syntax->entries);
      while(entry)
      {
         syntax = (GdsEditorSyntaxEntry *)entry->data;
         syntax->font = font;
         entry = entry->next;
      }
      entry = g_list_first(table->pattern->entries);
      while(entry)
      {
         pattern = (GdsEditorPatternEntry *)entry->data;
         pattern->font = font;
         entry = entry->next;
      }
      tables = tables->next;
   }
}

gchar *get_file_var(gchar *filename, const gchar var[])
{
   FILE *file;
   gchar varstring[256];
   gchar buffer[1024];
   gchar value[256];
   gchar stringer[] = "%s";
   g_snprintf(varstring, sizeof(varstring), "%s=%s", var, stringer);
   file = fopen(filename, "r");
   if(file == NULL)
   {
      return(NULL);
   }
   while(fgets(buffer, sizeof(buffer), file) != NULL)
   {
      if(strstr(buffer, var) != NULL)
      {
         sscanf(buffer, varstring, &value);
         fclose(file);
         return(value);
      }
   }
   return(NULL);
}

