#include "gdseditor.h"
#include "../gtkextext/gtkextext.h"
#include "gdsfile.h"
#include "gdstearaway.h"

static void gds_file_init(GdsFile *gds_file);
static void gds_file_class_init(GdsFileClass *klass);
GtkWidget* gds_file_new(gchar *filename);
static void gds_file_destroy(GtkObject *object);
static gint button_press_cb(GtkWidget *widget, GdkEventButton *event);
static gint key_press_cb(GtkWidget *widget, GdkEventKey *event, GdsFile *gds_file);
static gchar *make_spaces_string(gint spaces);

static GtkVBoxClass *parent_class = NULL;

GtkType gds_file_get_type(void)
{
   static GtkType gds_file_type=0;
   if(!gds_file_type)
   {
      static const GtkTypeInfo gds_file_info = 
      {	
         "GdsFile",
         sizeof(GdsFile),
         sizeof(GdsFileClass),
         (GtkClassInitFunc) gds_file_class_init,
         (GtkObjectInitFunc) gds_file_init,
         NULL,
         NULL,
         (GtkClassInitFunc)NULL,
      };
      gds_file_type = gtk_type_unique(GTK_TYPE_VBOX, &gds_file_info);
   }
   return(gds_file_type);
}

static void gds_file_class_init(GdsFileClass *klass)
{
   GtkObjectClass *object_class;
   object_class = (GtkObjectClass*)klass;
   parent_class = gtk_type_class(GTK_TYPE_HBOX);
   object_class->destroy = gds_file_destroy;
}

static void gds_file_init(GdsFile *gds_file)
{
   gds_file->filename = NULL;
   gds_file->tables = NULL;
   gds_file->props.over_ride = FALSE;
   gds_file->props.dir = NULL;
   gds_file->props.auto_indent = FALSE;
   gds_file->props.use_spaces = FALSE;
   gds_file->props.spaces = 0;
   gds_file->props.compiler = NULL;
   gds_file->props.debugger = NULL;
   gds_file->props.execution = NULL;
   gds_file->compile_id = 0;
   gds_file->compile_file = NULL;
   gds_file->debug_pid = 0;   
   gds_file->exec_pid = 0;
}

GtkWidget* gds_file_new(gchar *filename)
{
   GdsFile *gds_file;
   GtkWidget *label;
   GtkWidget *vbox;
   GtkWidget *hbox;
   GtkWidget *compile_window;
   GtkWidget *hscroll;
   gchar *str;
   
   gds_file = gtk_type_new(GDS_TYPE_FILE);
   gds_file_set_filename(gds_file, filename);

   if((str = strrchr(filename, '/')) != NULL) { str++; } else { str = filename; }
   gds_file->menu_item = gnome_stock_menu_item(GNOME_STOCK_PIXMAP_NEW, str);
   gtk_widget_show(gds_file->menu_item); 

   gds_file->tear_away = gds_tear_away_new(GTK_POS_LEFT);
   gtk_widget_show(gds_file->tear_away);
   
   gds_file->paned = gtk_vpaned_new();
   gtk_paned_set_gutter_size(GTK_PANED(gds_file->paned), 12);
   gds_tear_away_set_child(GDS_TEAR_AWAY(gds_file->tear_away), gds_file->paned);
   gtk_widget_show(gds_file->paned);

   vbox = gtk_vbox_new(FALSE, 0);
   gtk_widget_show(vbox);
   hbox = gtk_hbox_new(FALSE, 0);
   gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
   gtk_widget_show(hbox);

   gds_file->text = gds_editor_new();
   gtk_widget_set_style(gds_file->text, extext_style);
   gtk_widget_set_usize(GTK_WIDGET(gds_file->text), 550, 300);
   gtk_box_pack_start(GTK_BOX(hbox), gds_file->text, TRUE, TRUE, 0);
   gtk_widget_show(gds_file->text);
   gtk_extext_undo_set_max(GTK_EXTEXT(gds_file->text), general_preferences.undo_max);
   gds_file->vscroller = gtk_vscrollbar_new(GTK_EXTEXT(gds_file->text)->vadj);
   gtk_box_pack_start(GTK_BOX(hbox), gds_file->vscroller, FALSE, FALSE, 0);
   gtk_widget_show(gds_file->vscroller);
   gtk_signal_connect(GTK_OBJECT(gds_file->text), "button_press_event", GTK_SIGNAL_FUNC(button_press_cb), 0);
   gtk_signal_connect(GTK_OBJECT(gds_file->text), "key_press_event", GTK_SIGNAL_FUNC(key_press_cb), gds_file);
   hscroll = gtk_hscrollbar_new(GTK_EXTEXT(gds_file->text)->hadj);
   gtk_box_pack_start(GTK_BOX(vbox), hscroll, FALSE, TRUE, 0);
   gtk_widget_show(hscroll);
   gds_file->file_data = gtk_entry_new();
   gtk_box_pack_end(GTK_BOX(vbox), gds_file->file_data, FALSE, TRUE, 0);
   gtk_widget_show(gds_file->file_data);
   gtk_editable_set_editable(GTK_EDITABLE(gds_file->file_data), FALSE);
   gtk_paned_add1(GTK_PANED(gds_file->paned), vbox);

   gds_file->notebook = gtk_notebook_new();
   gtk_notebook_set_tab_pos(GTK_NOTEBOOK(gds_file->notebook), GTK_POS_BOTTOM);
   gtk_widget_show(gds_file->notebook);
   gtk_paned_add2(GTK_PANED(gds_file->paned), gds_file->notebook);
   
   label = gtk_label_new("Compile");
   gtk_widget_show(label);
   compile_window = gtk_scrolled_window_new(NULL, NULL);
   gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(compile_window), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
   gtk_widget_show(compile_window);
   gds_file->compile_list = gtk_clist_new(1);
   gtk_clist_set_selection_mode(GTK_CLIST(gds_file->compile_list), GTK_SELECTION_SINGLE);
   gtk_clist_set_shadow_type(GTK_CLIST(gds_file->compile_list), GTK_SHADOW_NONE);
   gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(compile_window), gds_file->compile_list);
   gtk_widget_show(gds_file->compile_list);
   gtk_notebook_append_page(GTK_NOTEBOOK(gds_file->notebook), compile_window, label);
   return GTK_WIDGET(gds_file);
}

static void gds_file_destroy(GtkObject *object)
{
   GdsFile *gds_file;
   g_return_if_fail(object != NULL);
   g_return_if_fail(GDS_IS_FILE(object));
   gds_file = GDS_FILE(object);
   gtk_widget_destroy(gds_file->menu_item);
   gtk_widget_destroy(gds_file->tear_away);
   g_free(gds_file->filename);
   g_free(gds_file->props.dir);
   g_free(gds_file->props.compiler);
   g_free(gds_file->props.debugger);
   g_free(gds_file->props.execution);
   gtk_widget_destroy(GTK_WIDGET(gds_file));
}

void gds_file_set_filename(GdsFile *gds_file, gchar *filename)
{
   g_return_if_fail(gds_file != NULL);
   g_return_if_fail(GDS_IS_FILE(gds_file));
   if(gds_file->filename)
      g_free(gds_file->filename);
   gds_file->filename = g_new(char, strlen(filename)+1);
   strcpy(gds_file->filename, filename);
}

void gds_file_set_title(GdsFile *gds_file, gchar *title)
{
   GtkWidget *parent;
   gchar *str;
   gint pos = 0;
   g_return_if_fail(gds_file != NULL);
   g_return_if_fail(GDS_IS_FILE(gds_file));
   parent = gds_file->menu_item->parent;
   
   if((str = strrchr(title, '/')) != NULL) { str++; } else { str = title; }
   pos = g_list_index(GTK_MENU_SHELL(parent)->children, gds_file->menu_item);
   gtk_widget_destroy(gds_file->menu_item);   

   gds_file->menu_item = gnome_stock_menu_item(GNOME_STOCK_PIXMAP_NEW, str);
   gtk_menu_shell_insert(GTK_MENU_SHELL(parent), gds_file->menu_item, pos);
   gtk_widget_show(gds_file->menu_item);

   gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(files_book), gds_file->tear_away, str);
   gtk_notebook_set_menu_label_text(GTK_NOTEBOOK(files_book), gds_file->tear_away, str);
   gds_tear_away_set_title(GDS_TEAR_AWAY(gds_file->tear_away), str);
}


static gint button_press_cb(GtkWidget *widget, GdkEventButton *event)
{
   GtkExText *text;
   gboolean handled = FALSE;
   gboolean shift_state;
   gboolean control_state;
   gint skip_size = 10;

   shift_state = event->state & GDK_SHIFT_MASK;
   control_state = event->state & GDK_CONTROL_MASK;  

   if(shift_state) skip_size *= 4;
   if(control_state) skip_size *= 8;  

   text = (GtkExText *)widget;

   if(!GTK_WIDGET_HAS_FOCUS(widget))
      gtk_widget_grab_focus(widget);

   if(event->type == GDK_BUTTON_PRESS && event->button == 3)
   {
      gtk_menu_popup(GTK_MENU(edit_popup_menu), NULL, NULL, NULL, NULL, event->button, event->time);
      handled = TRUE;
   }
   if(handled) gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "button_press_event");
   return(handled);
}

static gint key_press_cb(GtkWidget *widget, GdkEventKey *event, GdsFile *gds_file)
{
   GtkExText *text;
   GtkEditable *editable;
   gint curpos;
   gboolean handled = FALSE;
   gint i = 0;
   gchar *chars = NULL, *last_char = NULL, *spaces = NULL;
   gboolean extra_tab = FALSE;
   gboolean shift_state;
   gboolean control_state;
   gint key;
   
   shift_state = event->state & GDK_SHIFT_MASK;
   control_state = event->state & GDK_CONTROL_MASK;  
   key=event->keyval;
   
   text = (GtkExText *)widget;
   editable = (GtkEditable *)widget;
   curpos = gtk_extext_get_position(editable);
     
   spaces = make_spaces_string(gds_file->props.spaces);
   if(event->keyval == GDK_Tab || event->keyval == ' ')
   {
      if(editable->has_selection)
      {
         gint start = 0;
	 gint end = 0;
	 gint first_line = 0;
	 gint last_line = 0;
	 gint current_line = 0;
	 GtkExTextLineData *lineptr = NULL;
	 start = editable->selection_start_pos;
	 end = editable->selection_end_pos;
         gtk_extext_set_position(editable, start);
	 first_line = gtk_extext_get_line(text);
         gtk_extext_set_position(editable, end);
	 last_line = gtk_extext_get_line(text);
	 current_line = first_line;
         while(current_line <= last_line)
         {
            gtk_extext_set_column(text, 0);
            gtk_extext_set_line(text, current_line);
            chars = gtk_editable_get_chars(editable, editable->current_pos, editable->current_pos + gds_file->props.spaces);
            if(event->keyval == GDK_Tab && (control_state || shift_state))
            {
               if(!strcmp(spaces, chars))
                  gtk_editable_delete_text(editable, editable->current_pos, editable->current_pos + gds_file->props.spaces);
               else if(chars[0] == '\t')
                  gtk_editable_delete_text(editable, editable->current_pos, editable->current_pos + 1);
            }
            else if(event->keyval == GDK_Tab)
            {
               if(gds_file->props.use_spaces)
                  gtk_editable_insert_text(editable, spaces, strlen(spaces), &editable->current_pos);
               else
                  gtk_editable_insert_text(editable, "\t", 1, &editable->current_pos);
            }
            else if(event->keyval == ' ' && (control_state || shift_state))
            {
               if(chars[0] == ' ')
                  gtk_editable_delete_text(editable, editable->current_pos, editable->current_pos + 1);
            }
            else if(event->keyval == ' ')
            {
               gtk_editable_insert_text(editable, " ", 1, &editable->current_pos);
            }
            g_free(chars);
            current_line++;
         }	    
         lineptr = gtk_extext_get_line_data(text, first_line, NULL);
         start = lineptr->startpos;
         g_free(lineptr);
         lineptr = gtk_extext_get_line_data(text, last_line, NULL);
         end = lineptr->endpos-1;
         g_free(lineptr);
         gtk_extext_set_position(editable, end);
         gtk_editable_select_region(editable, start, end);
         handled = TRUE;
      }
      else if(event->keyval == GDK_Tab)
      {
         if(gds_file->props.use_spaces)
            gtk_editable_insert_text(editable, spaces, strlen(spaces), &curpos);
         else
            gtk_editable_insert_text(editable, "\t", 1, &curpos);
         handled = TRUE;
      }
   }
   else if(event->keyval == GDK_KP_Enter || event->keyval == GDK_Return)
   {
      if(gds_file->props.auto_indent)
      {
         gtk_extext_set_column(text, 0);
         last_char = gtk_editable_get_chars(editable, curpos-1, curpos);
         if(last_char && last_char[0] == '{') extra_tab = TRUE;
         chars = gtk_editable_get_chars(editable, editable->current_pos, curpos);
         if(chars)
         {
            for(i = 0; i < strlen(chars); i++)
            {
               if(chars[i] == '\t' || chars[i] == ' ') continue;
               else break;
            }
         }
      }
      gtk_editable_insert_text(editable, "\n", 1, &curpos);
      if(gds_file->props.auto_indent)
      {
         if(i) gtk_editable_insert_text(editable, chars, i, &editable->current_pos);
         if(extra_tab)
         {
            if(gds_file->props.use_spaces)
               gtk_editable_insert_text(editable, spaces, strlen(spaces), &editable->current_pos);
            else
               gtk_editable_insert_text(editable, "\t", 1, &editable->current_pos);
         }
         g_free(last_char);
         g_free(chars);
      }
      handled = TRUE;
   } 
   g_free(spaces);
   if(handled) gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_press_event");
   return(handled);
}

static gchar *make_spaces_string(gint spaces)
{
   gchar *string = NULL;
   if(spaces > 10) spaces = 10;
   else if(spaces < 0) spaces = 1;
   string = g_new(char, spaces+1);
   if(string)
   {
      memset(string, ' ', spaces);
      string[spaces] = '\0';
   }
   return(string);
}
