#include "gnomedirlist.h"
#include "../xpm/folder_closed.xpm"
#include "../xpm/folder_open.xpm"

/* function declarations */
static void gnome_dirlist_class_init(GnomeDirListClass *_class);
static void gnome_dirlist_init(GnomeDirList *dir_list);
GtkWidget *gnome_dirlist_new(void);
static void gnome_dirlist_destroy(GtkObject *object);
static void ctree_select_event(GtkCTree *tree, gint row, gint col, GdkEvent *event, GnomeDirList *dir_list);
static void ctree_unselect_event(GtkCTree *tree, gint row, gint col, GdkEvent *event, GnomeDirList *dir_list);
static void gnome_dirlist_fill_dir(GnomeDirList *dir_list, GtkCTreeNode *parent);
static void find_full_path(GnomeDirList *dir_list, GtkCTreeNode *node);
static void gnome_dirlist_set_dir(GnomeDirList *dir_list);
/* end function declarations */

static GtkWindowClass *parent_class = NULL;

GtkType gnome_dirlist_get_type(void)
{
   static GtkType dirlist_type=0;
   if(!dirlist_type)
   {
      static const GtkTypeInfo dirlist_info = 
      {	
         "GnomeDirList",
	 sizeof(GnomeDirList),
	 sizeof(GnomeDirListClass),
	 (GtkClassInitFunc) gnome_dirlist_class_init,
	 (GtkObjectInitFunc) gnome_dirlist_init,
	 NULL,
	 NULL,
	 (GtkClassInitFunc)NULL,
      };
      dirlist_type = gtk_type_unique(GTK_TYPE_WINDOW, &dirlist_info);
   }
   return(dirlist_type);
}

static void gnome_dirlist_class_init(GnomeDirListClass *_class)
{
   GtkObjectClass *object_class;
   object_class = (GtkObjectClass*)_class;
   parent_class = gtk_type_class(GTK_TYPE_WINDOW);
   object_class->destroy = gnome_dirlist_destroy;
}

static void gnome_dirlist_init(GnomeDirList *dir_list)
{
   dir_list->dirs = NULL;
   dir_list->entry = NULL; 
   dir_list->path[0] = '\0';
}

GtkWidget* gnome_dirlist_new()
{
   GnomeDirList *dir_list;
   GtkWidget *main_box;
   GtkWidget *util_box;
   GtkWidget *scrolled_window;
   GtkWidget *label;
   GtkWidget *hsep;

   dir_list = gtk_type_new(GNOME_TYPE_DIRLIST);

   gtk_container_set_border_width(GTK_CONTAINER(dir_list), 5);
   main_box = gtk_vbox_new(FALSE, 0);
   gtk_container_add(GTK_CONTAINER(dir_list), main_box);
   gtk_widget_show(main_box);

   util_box = gtk_hbox_new(FALSE, 0);
   gtk_box_pack_start(GTK_BOX(main_box), util_box, TRUE, TRUE, 5);
   gtk_widget_show(util_box);

   scrolled_window = gtk_scrolled_window_new(0, 0);
   gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
   gtk_widget_set_usize(scrolled_window, 200, 250);
   gtk_box_pack_start(GTK_BOX(util_box), scrolled_window, TRUE, TRUE, 0);
   gtk_widget_show(scrolled_window);

   dir_list->dirs = gtk_ctree_new(1, 0);
   gtk_clist_set_column_title(GTK_CLIST(dir_list->dirs), 0, "Directories");
   gtk_clist_set_selection_mode(GTK_CLIST(dir_list->dirs), GTK_SELECTION_SINGLE);
   gtk_clist_set_shadow_type(GTK_CLIST(dir_list->dirs), GTK_SHADOW_ETCHED_IN);
   gtk_clist_set_column_justification(GTK_CLIST(dir_list->dirs), 0, GTK_JUSTIFY_LEFT);
   gtk_clist_set_row_height(GTK_CLIST(dir_list->dirs), 16);
   gtk_clist_column_titles_show(GTK_CLIST(dir_list->dirs));
   gtk_ctree_set_expander_style(GTK_CTREE(dir_list->dirs), GTK_CTREE_EXPANDER_NONE);
   gtk_container_add(GTK_CONTAINER(scrolled_window), dir_list->dirs);
   gtk_signal_connect(GTK_OBJECT(dir_list->dirs), "select_row", GTK_SIGNAL_FUNC(ctree_select_event), dir_list);
   gtk_signal_connect(GTK_OBJECT(dir_list->dirs), "unselect_row", GTK_SIGNAL_FUNC(ctree_unselect_event), dir_list);
   gtk_widget_show(dir_list->dirs);

   util_box = gtk_hbox_new(FALSE, 0);
   gtk_box_pack_start(GTK_BOX(main_box), util_box, TRUE, TRUE, 0);
   gtk_widget_show(util_box);

   label = gtk_label_new("Path:");
   gtk_box_pack_start(GTK_BOX(util_box), label, FALSE, FALSE, 5);   
   gtk_widget_show(label);
   dir_list->entry = gtk_entry_new();
   gtk_box_pack_start(GTK_BOX(util_box), dir_list->entry, FALSE, TRUE, 5);   
   gtk_widget_show(dir_list->entry);
   
   hsep = gtk_hseparator_new();
   gtk_box_pack_start(GTK_BOX(main_box), hsep, TRUE, TRUE, 10);
   gtk_widget_show(hsep);

   util_box = gtk_hbox_new(FALSE, 0);
   gtk_box_pack_start(GTK_BOX(main_box), util_box, TRUE, TRUE, 0);
   gtk_widget_show(util_box);

   dir_list->cancel_button = gnome_stock_button(GNOME_STOCK_BUTTON_CANCEL);
   gtk_box_pack_end(GTK_BOX(util_box), dir_list->cancel_button, FALSE, FALSE, 5);
   gtk_widget_show(dir_list->cancel_button);   

   dir_list->ok_button = gnome_stock_button(GNOME_STOCK_BUTTON_OK);
   gtk_box_pack_end(GTK_BOX(util_box), dir_list->ok_button, FALSE, FALSE, 5);
   gtk_widget_show(dir_list->ok_button);   

   strncpy(dir_list->path, "/", sizeof(dir_list->path));
   dir_list->folder_closed = (GnomePixmap *)gnome_pixmap_new_from_xpm_d(folder_closed_xpm);
   dir_list->folder_open = (GnomePixmap *)gnome_pixmap_new_from_xpm_d(folder_open_xpm);
   gnome_dirlist_fill_dir(dir_list, NULL);
   gnome_dirlist_set_dir(dir_list);
   return GTK_WIDGET(dir_list);
}

static void gnome_dirlist_destroy(GtkObject *object)
{
   GtkWidget *dir_list;
   g_return_if_fail(object != NULL);
   g_return_if_fail(GNOME_IS_DIRLIST(object));
   dir_list = GTK_WIDGET(object);
   gtk_widget_destroy(dir_list);
}

static void gnome_dirlist_fill_dir(GnomeDirList *dir_list, GtkCTreeNode *parent)
{
   DIR *dir_file;
   struct dirent *dir;
   struct stat st;
   gchar filename[384];
   gchar up_all[] = ".";
   gchar up_one[] = "..";
   GtkCTreeNode *node;
   gchar *str;

   gtk_clist_freeze(GTK_CLIST(dir_list->dirs));
   dir_file = opendir(dir_list->path);
   if(dir_file != NULL)
   {
      while((dir = readdir(dir_file)) != NULL)
      {
         if(!strcmp(dir->d_name, up_one) || !strcmp(dir->d_name, up_all))
 	    continue;
         g_snprintf(filename, sizeof(filename), "%s/%s", dir_list->path, dir->d_name);
	 stat(filename, &st);
	 if(S_ISDIR(st.st_mode))
         {
            str = dir->d_name;
	    node = gtk_ctree_insert_node(GTK_CTREE(dir_list->dirs), parent, NULL, &str, 4, dir_list->folder_closed->pixmap, dir_list->folder_closed->mask, dir_list->folder_open->pixmap, dir_list->folder_open->mask, FALSE, FALSE);
         }	    
      }
   }
   else
   {
      g_print("The directory was not found!\n");
   }
   gtk_clist_thaw(GTK_CLIST(dir_list->dirs));
}


static void ctree_select_event(GtkCTree *tree, gint row, gint col, GdkEvent *event, GnomeDirList *dir_list)
{
   GtkCTreeNode *node;
   node = gtk_ctree_node_nth(tree, row);
   find_full_path(dir_list, node);
   if(event->type == GDK_2BUTTON_PRESS)
   {
      gnome_dirlist_fill_dir(dir_list, node);
      gtk_ctree_expand(GTK_CTREE(dir_list->dirs), node);
   }
   gnome_dirlist_set_dir(dir_list);
}

static void ctree_unselect_event(GtkCTree *tree, gint row, gint col, GdkEvent *event, GnomeDirList *dir_list)
{
   strncpy(dir_list->path, "/", sizeof(dir_list->path));
   gnome_dirlist_set_dir(dir_list);
}

static void find_full_path(GnomeDirList *dir_list, GtkCTreeNode *node)
{
   gchar *ptr;
   gchar buf[384];
   gchar *insert;
   gint length;
   g_return_if_fail(dir_list != NULL);
   g_return_if_fail(GNOME_IS_DIRLIST(dir_list));

   insert = buf + sizeof(buf)-1;
   *insert = '\0';
   insert--;
   
   while(node)
   {
      gtk_ctree_node_get_pixtext(GTK_CTREE(dir_list->dirs), node, 0, &ptr, NULL, NULL, NULL);
      for(length = strlen(ptr)-1; length >= 0; length--)
      {
	 *insert = ptr[length];
         insert--;
      }
      *insert = '/';
      insert--;
      node = GTK_CTREE_ROW(node)->parent;
   }
   insert++;
   strcpy(dir_list->path, insert);
}

static void gnome_dirlist_set_dir(GnomeDirList *dir_list)
{
   g_return_if_fail(dir_list != NULL);
   g_return_if_fail(GNOME_IS_DIRLIST(dir_list));
   gtk_entry_set_text(GTK_ENTRY(dir_list->entry), dir_list->path);
}

gchar *gnome_dirlist_get_dir(GnomeDirList *dir_list)
{
   g_return_val_if_fail(dir_list != NULL, NULL);
   g_return_val_if_fail(GNOME_IS_DIRLIST(dir_list), NULL);
   return(dir_list->path);
}
