#include "gnomegotoline.h"

/* function declarations */
static void gnome_goto_line_dialog_class_init(GnomeGotoLineDialogClass *klass);
static void gnome_goto_line_dialog_init(GnomeGotoLineDialog *gtl);
GtkWidget *gnome_goto_line_dialog_new(void);
static void gnome_goto_line_dialog_destroy(GtkObject *object);
/* end function declarations */

static GtkWindowClass *parent_class = NULL;

GtkType gnome_goto_line_dialog_get_type(void)
{
   static GtkType goto_line_type=0;
   if(!goto_line_type)
   {
      static const GtkTypeInfo goto_line_info = 
      {	
         "GnomeGotoLineDialog",
         sizeof(GnomeGotoLineDialog),
         sizeof(GnomeGotoLineDialogClass),
         (GtkClassInitFunc) gnome_goto_line_dialog_class_init,
         (GtkObjectInitFunc) gnome_goto_line_dialog_init,
         NULL,
         NULL,
         (GtkClassInitFunc)NULL,
      };
      goto_line_type = gtk_type_unique(GTK_TYPE_WINDOW, &goto_line_info);
   }
   return(goto_line_type);
}

static void gnome_goto_line_dialog_class_init(GnomeGotoLineDialogClass *klass)
{
   GtkObjectClass *object_class;
   object_class = (GtkObjectClass*)klass;
   parent_class = gtk_type_class(GTK_TYPE_WINDOW);
   object_class->destroy = gnome_goto_line_dialog_destroy;
}

static void gnome_goto_line_dialog_init(GnomeGotoLineDialog *gtl)
{
   gtl->line_entry = NULL;
   gtl->goto_button = NULL;
}

GtkWidget* gnome_goto_line_dialog_new()
{
   GnomeGotoLineDialog *gtl;
   GtkWidget *main_box;
   GtkWidget *util_box;
   GtkWidget *label;
   GtkWidget *hsep;

   gtl = gtk_type_new(GNOME_TYPE_GOTO_LINE_DIALOG);
   gtk_container_set_border_width(GTK_CONTAINER(gtl), 5);
   gtk_window_set_title(GTK_WINDOW(gtl), "Goto Line...");
   
   main_box = gtk_vbox_new(FALSE, 0);
   gtk_container_add(GTK_CONTAINER(gtl), main_box);
   gtk_widget_show(main_box);

   util_box = gtk_hbox_new(FALSE, 0);
   gtk_box_pack_start(GTK_BOX(main_box), util_box, TRUE, TRUE, 0);
   gtk_widget_show(util_box);

   label = gtk_label_new("Line #: ");
   gtk_box_pack_start(GTK_BOX(util_box), label, FALSE, FALSE, 0);   
   gtk_widget_show(label);

   gtl->line_entry = gtk_entry_new();
   gtk_box_pack_start(GTK_BOX(util_box), gtl->line_entry, TRUE, TRUE, 0);   
   gtk_widget_show(gtl->line_entry);

   hsep = gtk_hseparator_new();
   gtk_box_pack_start(GTK_BOX(main_box), hsep, TRUE, TRUE, 5);
   gtk_widget_show(hsep);

   util_box = gtk_hbox_new(FALSE, 0);
   gtk_box_pack_start(GTK_BOX(main_box), util_box, TRUE, TRUE, 0);
   gtk_widget_show(util_box);

   gtl->cancel_button = gnome_stock_button(GNOME_STOCK_BUTTON_CANCEL);
   gtk_box_pack_end(GTK_BOX(util_box), gtl->cancel_button, FALSE, FALSE, 5);
   gtk_widget_show(gtl->cancel_button);   

   gtl->goto_button = gnome_pixmap_button(gnome_stock_pixmap_widget_at_size(GTK_WIDGET(gtl), GNOME_STOCK_PIXMAP_SEARCH, 20, 20), "Goto Line");
   gtk_box_pack_end(GTK_BOX(util_box), gtl->goto_button, FALSE, FALSE, 5);
   gtk_widget_show(gtl->goto_button);   

   return GTK_WIDGET(gtl);
}

static void gnome_goto_line_dialog_destroy(GtkObject *object)
{
   GtkWidget *gtl;
   g_return_if_fail(object != NULL);
   g_return_if_fail(GNOME_IS_GOTO_LINE_DIALOG(object));
   gtl = GTK_WIDGET(object);
   gtk_widget_destroy(gtl);
}

gint gnome_goto_line_dialog_get_line(GnomeGotoLineDialog *gtl)
{
   gint line_num;
   gchar *line_txt;

   g_return_val_if_fail(gtl != NULL, -1);
   g_return_val_if_fail(GNOME_IS_GOTO_LINE_DIALOG(gtl), -1);

   line_txt = gtk_entry_get_text(GTK_ENTRY(gtl->line_entry));
   if(!strlen(line_txt))
      return(-1);
   if(!sscanf(line_txt, "%d", &line_num))
      return(-1);
   if(line_num < 1)
      return(-1);
   return(line_num-1);
}

